# Author: Martin Kristiansen

from itertools import combinations
import itertools
import math


class CostBenefit(object):
    '''
    Experimental class for cost-benefit calculations.

    Currently including allocation schemes based on "cooperative game theory"
    '''


    def __init__(self):
        """Collect value-functions for each player in the expansion-game

        Parameters
        ----------


        Creat CostBenefit object:
        """
        self.players = None
        self.coalitions = None
        self.valueFunction = None
        self.payoff = None

    def power_set(self,List):
        """
        function to return the powerset of a list, i.e. all possible subsets ranging
        from length of one, to the length of the larger list
        """

        subs = [list(j) for i in range(len(List)) for j in combinations(List, i+1)]
        return subs

    def nCr(self,n,r):
        '''
        calculate the binomial coefficient, i.e. how many different possible
        subsets can be made from the larger set n
        '''
        f = math.factorial
        return f(n) / f(r) / f(n-r)


    def gameSetup(self, grid_data):
        self.players = grid_data.node.area.unique().tolist()
        self.coalitions = self.power_set(self.players)


    def getBinaryCombinations(self,num):
        '''
        Returns a sequence of different combinations 1/0 for a number of
        decision variables. E.g. three cable investments;
        (0,0,0), (1,0,0), (0,1,0), and so on.
        '''
        combinations = list(itertools.product([0,1], repeat=num))
        return combinations


    def gameShapleyValue(self,player_list, values):
        '''compute the Shapley Value from cooperative game theory

       Parameters:
        ===========
        player_list: list of all players in the game
        values: characteristic function for each subset of N players, i.e.
                possible coaltions/cooperations among players.

        Returns the Shapley value, i.e. a fair cost/benefit allocation based
        on the average marginal contribution from each player.

        '''

        if type(values) is not dict:
            raise TypeError("characteristic function must be a dictionary")
        for key in values:
            if len(str(key)) == 1 and type(key) is not tuple:
                values[(key,)] = values.pop(key)
            elif type(key) is not tuple:
                raise TypeError("key must be a tuple")
        for key in values:
            sortedkey = tuple(sorted(list(key)))
            values[sortedkey] = values.pop(key)

        player_list = max(values.keys(), key=lambda key: len(key))
        for coalition in self.power_set(player_list):
            if tuple(sorted(list(coalition))) not in sorted(values.keys()):
                raise ValueError("characteristic function must be the power set")

        payoff_vector = {}
        n = len(player_list)
        for player in player_list:
            weighted_contribution = 0
            for coalition in self.power_set(player_list):
                if coalition:  # If non-empty
                    k = len(coalition)
                    weight = 1/(self.nCr(n,k)*k)
                    t = tuple(p for p in coalition if p != player)
                    weighted_contribution += weight * (values[tuple(coalition)]
                                                       - values[t])
            payoff_vector[player] = weighted_contribution

        return payoff_vector

    def gameIsMonotone(self, values):
        '''
        Returns true if the game/valueFunction is monotonic.
        A game G = (N, v) is monotonic if it satisfies the value function
        of a subset is less or equal then the value function from its
        union set:
        v(C_2) \geq v(C_1) for all C_1 \subseteq C_2
        '''
        return not any([set(p1) <= set(p2) and values[p1] > values[p2]
            for p1, p2 in itertools.permutations(values.keys(), 2)])


    def gameIsSuperadditive(self, values):
        '''
        Returns true if the game/valueFunction is superadditive.
        A characteristic function game G = (N, v) is superadditive
        if it the sum of two coalitions/subsets gives a larger value than the
        individual sum:
        v(C_1 \cup C_2) \geq v(C_1) +  v(C_2) for
        all C_1, C_2 \subseteq 2^{\Omega} such that C_1 \cap C_2
        = \emptyset.
        '''
        sets = values.keys()
        for p1, p2 in itertools.combinations(sets, 2):
            if not (set(p1) & set(p2)):
                union = tuple(sorted(set(p1) | set(p2)))
                if values[union] < values[p1] + values[p2]:
                    return False
        return True

    def gamePayoffIsEfficient(self, player_list, values, payoff_vector):
        '''
        Return `true if the payoff vector is efficient. A payoff vector v is
        efficient if the sum of payments equal the total value provided by a
        set of players.
        \sum_{i=1}^N \lambda_i = v(\Omega);
        '''
        pl = tuple(sorted(list(player_list)))
        return sum(payoff_vector.values()) == values[pl]

    def gamePayoffHasNullplayer(self, player_list, values, payoff_vector):
        '''
        Return true if the payoff vector possesses the nullplayer property.
        A payoff vector v has the nullplayer property if there exists
        an i such that v(C \cup i) = v(C) for all C \in 2^{\Omega}
        then, \lambda_i = 0. In other words: if a player does not
        contribute to any coalition then that player should receive no payoff.
        '''
        for player in player_list:
            results = []
            for coalit in values:
                if player in coalit:
                    t = tuple(sorted(set(coalit) - {player}))
                    results.append(values[coalit] == values[t])
            if all(results) and payoff_vector[player] != 0:
                return False
        return True

    def gamePayoffIsSymmetric(self, values, payoff_vector):
        '''
        Returns true if the resulting payoff vector possesses the symmetry property.
        A payoff vector possesses the symmetry property if players with equal
        marginal contribution receives the same payoff:
        v(C \cup i) = v(C \cup j) for all
        C \in 2^{\Omega} \setminus \{i,j\}, then x_i = x_j.
        '''
        sets = values.keys()
        element = [i for i in sets if len(i) == 1]
        for c1, c2 in itertools.combinations(element, 2):
            results = []
            for m in sets:
                junion = tuple(sorted(set(c1) | set(m)))
                kunion = tuple(sorted(set(c2) | set(m)))
                results.append(values[junion] == values[kunion])
            if all(results) and payoff_vector[c1[0]] != payoff_vector[c2[0]]:
                return False
        return True



# -*- coding: utf-8 -*-
'''
Module for PowerGAMA constants
'''

baseS = 100.0e6
'''Per unit base value for power in W (100 MW)'''

baseV = 400.0e3
'''Per unit base value for voltage in V (400 kV)'''

loadshedcost = 1000.0
'''Penalty (/MWh) for load shedding'''

hoursperyear = 8760.0
'''Hours per year (365*24 = 8760)'''

MWh_per_GWh = 1000.0
'''Conversion factor from GWh to MWh'''

baseAngle = 1
'''Base value for voltage angle'''

# Derived quantities
#baseZ = baseV**2/baseS
baseMVA = baseS*1e-6

flexload_outside_cost = 1000.0
'''(Very high) storage value for flexible demand outside flexibility range'''



# -*- coding: utf-8 -*-
"""
PowerGAMA module containing plotting functions
"""

import math

import branca.colormap
import folium
import geopandas
import numpy as np
import pandas as pd
import shapely
from folium.plugins.overlapping_marker_spiderfier import OverlappingMarkerSpiderfier
from folium.plugins.treelayercontrol import TreeLayerControl


def plotMap(
    pg_data,
    pg_res=None,
    filename=None,
    nodetype=None,
    branchtype=None,
    filter_node=None,
    filter_branch=None,
    branch_capacity_width=True,
    timeMaxMin=None,
    spread_nodes_r=None,
    layer_control=True,
    fit_bound=True,
    scale_radius=1,
    res_types=None,
    **kwargs,
):
    """
    Plot PowerGAMA data/results on map

    Parameters
    ==========
    pg_data : powergama.GridData
        powergama data object
    pg_res : powergama.Results
        powergama results object
    filename : str
        name of output file (html)
    nodetype : str ('nodalprice','loadshedding','area','curtailed_res') or None (default)
        how to colour nodes
    branchtype : str ('utilisation','sensitivity','flow','capacity','type') or None (default)
        how to colour branches
    filter_node : list
        max/min value used for colouring nodes (e.g. nodalprice)
    filter_branch : list
        max/min value used for colouring branches (e.g. utilisation)
    branch_capacity_width : boolean
        scale branch width according to capacity, True (default) or False
    timeMaxMin : [min,max]
        time interval used when showing simulation results
    spread_nodes_r : float (degrees)
        radius (degrees) of circle on which overlapping nodes are
        spread (use eg 0.04)
    fit_bound : boolean
        whether to adjust bounds to fit the grid
    scale_radius : float
        scaling factor to adjust size of loads and generators
    res_types : list or None
        list of generator types considered RES
    kwargs : arguments passed on to folium.Map(...)
    """

    if nodetype not in [None, "area", "type", "nodalprice", "loadshedding", "curtailed_res"]:
        raise ValueError(f"Unknown nodetype: {nodetype}")
    if branchtype not in [None, "type", "capacity", "flow", "utilisation", "sensitivity"]:
        raise ValueError(f"Unknown branchtype: {branchtype}")

    cmSet1 = branca.colormap.linear.Set1_03

    # Add geographic information to branches and generators/consumers
    branch = pg_data.branch.copy()
    dcbranch = pg_data.dcbranch.copy()
    node = pg_data.node.copy()
    generator = pg_data.generator.copy()
    consumer = pg_data.consumer.copy()

    if spread_nodes_r is not None:
        # spread out nodes lying on top of each other
        coords = node[["lat", "lon"]]
        dupl_coords = pd.DataFrame()
        dupl_coords["cumcount"] = coords.groupby(["lat", "lon"]).cumcount()
        dupl_coords["count"] = coords.groupby(["lat", "lon"])["lon"].transform("count")
        for i in node.index:
            n_sum = dupl_coords.loc[i, "count"]
            if n_sum > 1:
                # there are more nodes with the same coordinates
                n = dupl_coords.loc[i, "cumcount"]
                theta = 2 * math.pi / n_sum
                node.loc[i, "lat"] += spread_nodes_r * math.cos(n * theta)
                node.loc[i, "lon"] += spread_nodes_r * math.sin(n * theta)
        # node[['lat','lon']] = coords

    # Careful! Merge may change the order
    branch = branch.merge(node[["id", "lat", "lon"]], how="left", left_on="node_from", right_on="id")
    branch = branch.merge(node[["id", "lat", "lon"]], how="left", left_on="node_to", right_on="id")
    dcbranch = dcbranch.merge(node[["id", "lat", "lon"]], how="left", left_on="node_from", right_on="id")
    dcbranch = dcbranch.merge(node[["id", "lat", "lon"]], how="left", left_on="node_to", right_on="id")
    branch_gen = pd.DataFrame()
    if ("gen_lat" in generator.columns) and ("gen_lon" in generator.columns):
        latlon_from_node = generator.merge(node[["id", "lat", "lon"]], how="left", left_on="node", right_on="id")
        generator["lat"] = generator["gen_lat"].fillna(latlon_from_node["lat"])
        generator["lon"] = generator["gen_lon"].fillna(latlon_from_node["lon"])
        branch_gen = pd.DataFrame(index=generator.index)
        branch_gen["lat_x"] = generator["gen_lat"]
        branch_gen["lat_y"] = latlon_from_node["lat"]
        branch_gen["lon_x"] = generator["gen_lon"]
        branch_gen["lon_y"] = latlon_from_node["lon"]
        # keep only those where generator lat/lon have been specified, and differs from node
        m_remove = (branch_gen.isna().any(axis=1)) | (
            (branch_gen["lat_x"] == branch_gen["lat_y"]) & (branch_gen["lon_x"] == branch_gen["lon_y"])
        )
        branch_gen = branch_gen[~m_remove]
    else:
        generator = generator.merge(node[["id", "lat", "lon"]], how="left", left_on="node", right_on="id")
    consumer = consumer.merge(node[["id", "lat", "lon"]], how="left", left_on="node", right_on="id")
    gentypes_all = pg_data.getAllGeneratorTypes()
    areas = pg_data.getAllAreas()
    node = node.reset_index().rename(columns={"index": "index_orig"})
    node = node.merge(pd.DataFrame(areas).reset_index(), how="left", left_on="area", right_on=0).rename(
        columns={"index": "area_ind"}
    )
    node = node.set_index("index_orig")
    # node.sort_index(inplace=True)
    node["area_ind"] = 0.5 + node["area_ind"] % 10

    node_value_col = None
    node_fields = []
    branch_value_col = None
    branch_fields = []

    if pg_res is not None:
        # if not specified, use entire time range
        if timeMaxMin is None:
            timeMaxMin = [pg_data.timerange[0], pg_data.timerange[-1] + 1]

        node_fields.append("nodalprice")
        node["nodalprice"] = pg_res.getAverageNodalPrices(timeMaxMin)
        node["loadshedding"] = pg_res.getLoadsheddingPerNode(average=True)
        node_fields.append("loadshedding")
        if nodetype == "curtailed_res":
            if res_types is None:
                raise ValueError("You must specify parameter 'res_types'")
            df_spill = pd.DataFrame()
            df_spill["node"] = pg_data.generator["node"]
            df_spill["type"] = pg_data.generator["type"]
            df_spill["curtailed_abs"] = pg_res.db.getResultGeneratorSpilledSums(timeMaxMin)
            df_spill["output"] = pg_res.db.getResultGeneratorPowerSum(timeMaxMin)
            # TODO replace hardcoded generator type by something better. This will nto always work:
            df_spill = df_spill[df_spill["type"].isin(res_types)].groupby("node").sum()
            df_spill["curtailed"] = df_spill["curtailed_abs"] / (df_spill["curtailed_abs"] + df_spill["output"])
            df_spill = pg_data.node.merge(df_spill, left_on="id", right_index=True, how="left")

            # gen_spilled = pg_res.db.getResultGeneratorSpilledSums(timeMaxMin)
            # gen_output = pg_res.db.getResultGeneratorPowerSum(timeMaxMin)
            # df_spill = pd.DataFrame(index=pg_data.generator["node"], data={"curtailed_abs":gen_spilled,"output":gen_output})
            # df_spill = df_spill.groupby("node").sum()
            # df_spill["curtailed"] = df_spill["curtailed_abs"]/(df_spill["curtailed_abs"]+df_spill["output"])
            # df_spill = pg_data.node.merge(df_spill,left_on="id",right_index=True,how="left")
            node["curtailed_res"] = df_spill["curtailed"].fillna(0)
            node_fields.append("curtailed_res")

        branch_sensitivity = pg_res.getAverageBranchSensitivity(timeMaxMin)
        branch_utilisation = pg_res.getAverageUtilisation(timeMaxMin)
        br_ind = pg_data.getIdxBranchesWithFlowConstraints()
        branch["flow12"] = pg_res.getAverageBranchFlows(timeMaxMin)[0]
        branch["flow21"] = pg_res.getAverageBranchFlows(timeMaxMin)[1]
        branch["flow"] = pg_res.getAverageBranchFlows(timeMaxMin)[2]  # avg of abs value
        branch["utilisation"] = branch_utilisation
        branch["sensitivity"] = 0.0
        branch.loc[branch.index.isin(br_ind), "sensitivity"] = branch_sensitivity
        # DC branch utilisation:
        if dcbranch.shape[0] > 0:
            dcbranch_utilisation = pg_res.getAverageUtilisation(timeMaxMin, branchtype="dc")
            dcbranch_sensitivity = pg_res.getAverageBranchSensitivity(timeMaxMin, branchtype="dc")
            dcbr_ind = pg_data.getIdxDcBranchesWithFlowConstraints()
            dcbranch["flow12"] = pg_res.getAverageBranchFlows(timeMaxMin, branchtype="dc")[0]
            dcbranch["flow21"] = pg_res.getAverageBranchFlows(timeMaxMin, branchtype="dc")[1]
            dcbranch["flow"] = pg_res.getAverageBranchFlows(timeMaxMin, branchtype="dc")[2]
            dcbranch["utilisation"] = 0.0
            dcbranch.loc[dcbranch.index.isin(dcbr_ind), "utilisation"] = dcbranch_utilisation
            dcbranch["sensitivity"] = 0.0
            dcbranch.loc[dcbranch.index.isin(dcbr_ind), "sensitivity"] = dcbranch_sensitivity
        branch_fields = ["utilisation", "flow12", "flow21", "flow", "sensitivity"]

    m = folium.Map(location=[node["lat"].median(), node["lon"].median()], **kwargs)

    # Node colouring
    node_cbar_caption = {"nodalprice": "Nodal price", "loadshedding": "Load shedding", "curtailed_res": "Curtailed RES"}
    if nodetype is None:
        pass
    elif nodetype in ["nodalprice", "loadshedding", "curtailed_res"]:
        node_value_col = nodetype
        if filter_node is None:
            filter_node = [node[node_value_col].min(), node[node_value_col].max()]
        cm_node = branca.colormap.LinearColormap(["green", "yellow", "red"], vmin=filter_node[0], vmax=filter_node[1])
        cm_node.caption = node_cbar_caption[nodetype]
        m.add_child(cm_node)
    elif nodetype == "area":
        node_value_col = "area_ind"
        val_max = node[node_value_col].max()
        cm_node = cmSet1.scale(0, val_max).to_step(10)
    elif nodetype == "type":
        pass
    else:
        raise ValueError(f"Unknown nodetype parameter: {nodetype}")

    # Branch colouring
    if branchtype is None:
        pass
    elif branchtype == "utilisation":
        branch_value_col = "utilisation"
        if filter_branch is None:
            filter_branch = [0, 1]
        cm_branch = branca.colormap.LinearColormap(
            ["green", "yellow", "red"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch utilisation"
        m.add_child(cm_branch)
    elif branchtype == "sensitivity":
        branch_value_col = "sensitivity"
        if filter_branch is None:
            filter_branch = [branch[branch_value_col].min(), branch[branch_value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch capacity sensitivity"
        m.add_child(cm_branch)
    elif branchtype == "flow":
        branch_value_col = "flow"
        if filter_branch is None:
            filter_branch = [branch[branch_value_col].min(), branch[branch_value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch flow (abs value)"
        m.add_child(cm_branch)
    elif branchtype == "capacity":
        branch_value_col = "capacity"
        if filter_branch is None:
            filter_branch = [branch[branch_value_col].min(), branch[branch_value_col].max()]
        cm_branch = branca.colormap.LinearColormap(
            ["red", "yellow", "green"], vmin=filter_branch[0], vmax=filter_branch[1]
        )
        cm_branch.caption = "Branch capacity"
        m.add_child(cm_branch)
    elif branchtype == "type":
        pass
    else:
        raise Exception("Invalid branchtype argument")

    # Styling
    cap_scale = 500
    radius_default = 4
    radius_max = 20
    radius_min = 2
    weight_min = 1
    weight_max = 10
    fill_opacity = 0.7

    def style_node(feature):
        style = {}
        if node_value_col:
            style["fillColor"] = cm_node(feature["properties"][node_value_col])
            style["fillOpacity"] = 1
        return style

    def style_branch(feature):
        style = {"color": "black", "weight": 2}
        if branch_capacity_width:
            style["weight"] = np.clip(feature["properties"]["capacity"] / cap_scale, a_min=weight_min, a_max=weight_max)
        if branchtype == "type":
            style["color"] = "red"
        elif branch_value_col:
            style["color"] = cm_branch(feature["properties"][branch_value_col])
        return style

    def style_dcbranch(feature):
        style = {"color": "black", "weight": 2}
        if branch_capacity_width:
            style["weight"] = np.clip(feature["properties"]["capacity"] / cap_scale, a_min=weight_min, a_max=weight_max)
        if branchtype == "type":
            style["color"] = "blue"
        elif branch_value_col:
            style["color"] = cm_branch(feature["properties"][branch_value_col])
        return style

    def style_generator(feature):
        style = {
            "color": "green",
            # "fillColor": "green",
            "radius": np.clip(
                np.sqrt(feature["properties"]["pmax"] / cap_scale) * scale_radius, a_min=radius_min, a_max=radius_max
            ),
        }
        return style

    def style_genbranch(feature):
        style = {
            "color": "green",
            "weight": 1,
        }
        return style

    def style_consumers(feature):
        style = {
            "radius": np.clip(
                np.sqrt(feature["properties"]["demand_avg"] / cap_scale) * scale_radius,
                a_min=radius_min,
                a_max=radius_max,
            )
        }
        return style

    # Nodes:
    map_nodes = geopandas.GeoDataFrame(
        node, geometry=geopandas.points_from_xy(node["lon"], node["lat"]), crs="EPSG:4326"
    )
    m_nodes = folium.GeoJson(
        map_nodes,
        style_function=style_node,
        marker=folium.CircleMarker(
            fill=True, fill_opacity=fill_opacity, radius=radius_default, weight=1, color="black"
        ),
        tooltip=folium.GeoJsonTooltip(fields=["id", "area", "zone"] + node_fields),
        name="Nodes",
    ).add_to(m)

    # Branches:
    line_geometry = [
        shapely.LineString([(z[0], z[1]), (z[2], z[3])])
        for z in zip(branch["lon_x"], branch["lat_x"], branch["lon_y"], branch["lat_y"])
    ]
    map_branches = geopandas.GeoDataFrame(branch, geometry=line_geometry, crs="EPSG:4326")
    m_acbranches = folium.GeoJson(
        map_branches,
        style_function=style_branch,
        tooltip=folium.GeoJsonTooltip(fields=["node_from", "node_to", "capacity", "reactance"] + branch_fields),
        name="AC branches",
    ).add_to(m)
    dcline_geometry = [
        shapely.LineString([(z[0], z[1]), (z[2], z[3])])
        for z in zip(dcbranch["lon_x"], dcbranch["lat_x"], dcbranch["lon_y"], dcbranch["lat_y"])
    ]
    if dcbranch.empty:
        m_dcbranches = None
        br_all = branch
    else:
        map_dcbranches = geopandas.GeoDataFrame(dcbranch, geometry=dcline_geometry, crs="EPSG:4326")
        m_dcbranches = folium.GeoJson(
            map_dcbranches,
            style_function=style_dcbranch,
            tooltip=folium.GeoJsonTooltip(fields=["node_from", "node_to", "capacity"] + branch_fields),
            name="DC branches",
        ).add_to(m)
        # branch flow direction marker:
        br_all = pd.concat([branch, dcbranch], axis=0, ignore_index=True)
    branch_flowmarker = pd.DataFrame(index=br_all.index)
    if pg_res is not None:
        for ind, n in br_all.iterrows():
            if n["flow12"] + n["flow21"] == 0:
                d = 0.5
            else:
                d = min(0.9, max(0.1, n["flow12"] / (n["flow12"] + n["flow21"])))
            (flow_lat, flow_lon) = _pointBetween((n["lat_x"], n["lon_x"]), (n["lat_y"], n["lon_y"]), weight=d)
            branch_flowmarker.loc[ind, "lat"] = flow_lat
            branch_flowmarker.loc[ind, "lon"] = flow_lon

        map_flowmarker = geopandas.GeoDataFrame(
            branch_flowmarker,
            geometry=geopandas.points_from_xy(branch_flowmarker["lon"], branch_flowmarker["lat"]),
            crs="EPSG:4326",
        )
        m_flowmarker = folium.GeoJson(
            map_flowmarker,
            marker=folium.CircleMarker(fill=True, fill_opacity=0.5, radius=radius_default, weight=1, color="gray"),
            name="Flow marker",
        ).add_to(m)
    # Generators:
    map_generators = geopandas.GeoDataFrame(
        generator, geometry=geopandas.points_from_xy(generator["lon"], generator["lat"]), crs="EPSG:4326"
    )
    if branch_gen.shape[0] > 0:
        genline_geometry = [
            shapely.LineString([(z[0], z[1]), (z[2], z[3])])
            for z in zip(branch_gen["lon_x"], branch_gen["lat_x"], branch_gen["lon_y"], branch_gen["lat_y"])
        ]
        map_genbranches = geopandas.GeoDataFrame(branch_gen, geometry=genline_geometry, crs="EPSG:4326")
    m_generators = {}
    for gentype in gentypes_all:
        m_generators[gentype] = folium.FeatureGroup(name=gentype).add_to(m)
        folium.GeoJson(
            map_generators[map_generators["type"] == gentype],
            marker=folium.CircleMarker(fill=True, fillOpacity=0.7, weight=1, fill_color="green"),
            style_function=style_generator,
            tooltip=folium.GeoJsonTooltip(fields=["node", "desc", "type", "pmax", "inflow_ref"]),
            # popup=folium.GeoJsonPopup(fields=["node", "desc", "type","pmax","inflow_ref"]),
            name=gentype,
        ).add_to(m_generators[gentype])
        if branch_gen.shape[0] > 0:
            genbranch_indices = [
                i for i in map_generators[map_generators["type"] == gentype].index if i in map_genbranches.index
            ]
            folium.GeoJson(
                map_genbranches.loc[genbranch_indices],
                style_function=style_genbranch,
                name=f"{gentype}:connection",
            ).add_to(m_generators[gentype])
    # Consumers:
    map_consumers = geopandas.GeoDataFrame(
        consumer, geometry=geopandas.points_from_xy(consumer["lon"], consumer["lat"]), crs="EPSG:4326"
    )
    m_consumers = folium.GeoJson(
        map_consumers,
        marker=folium.CircleMarker(fill=True, color="black", fill_color="yellow", fill_opacity=0.5, weight=1),
        style_function=style_consumers,
        tooltip=folium.GeoJsonTooltip(fields=["node", "demand_avg", "demand_ref"]),
        name="Consumers",
    ).add_to(m)

    if fit_bound:
        sw = node[["lat", "lon"]].min().values.tolist()
        ne = node[["lat", "lon"]].max().values.tolist()
        m.fit_bounds([sw, ne])
    if layer_control:
        branch_children = [
            {"label": "AC", "layer": m_acbranches},
            {"label": "DC", "layer": m_dcbranches},
        ]
        if pg_res:
            branch_children.append({"label": "flow markers", "layer": m_flowmarker})
        overlay_tree = {
            "label": "powergama",
            "select_all_checkbox": True,
            "children": [
                {"label": "Nodes", "layer": m_nodes},
                {
                    "label": "Branches",
                    "select_all_checkbox": True,
                    "children": branch_children,
                },
                {"label": "Consumers", "layer": m_consumers},
                {
                    "label": "Generators",
                    "select_all_checkbox": True,
                    "children": [{"label": gentype, "layer": m_generators[gentype]} for gentype in gentypes_all],
                },
            ],
        }

        TreeLayerControl(overlay_tree=overlay_tree).add_to(m)

    # TODO: This does not seem to work with GeoJson data markers
    oms = OverlappingMarkerSpiderfier()  # keep_spiderfied=True, nearby_distance=50, leg_weight=2.0)
    oms.add_to(m)

    if filename:
        print(f"Saving map to file {filename}")
        m.save(filename)

    return m


def _pointBetween(nodeA, nodeB, weight, mercator=True):
    """computes coords on the line between two points

     (lat,lon) = pointBetween(self,lat1,lon1,lat2,lon2,d)

    Parameters
    ----------
        nodeA: dublet (lat,lon)
            latitude/longitude of nodeA (degrees)
        nodeB: dublet (lat,lon)
            latitude/longitude of nodeB (degrees)
        weight: double
            weight=0 is node A, 0.5 is halfway between, and 1 is node B
        mercator: boolean
            True if getting point that looks right in mercator projection

    Returns
    -------
        (lat,lon): dublet
            coordinates of the point inbetween nodeA and nodeB (degrees)


     ref: http://www.movable-type.co.uk/scripts/latlong.html
    """
    lat1 = nodeA[0]
    lon1 = nodeA[1]
    lat2 = nodeB[0]
    lon2 = nodeB[1]
    if mercator:
        return mercator_interpolate(lat1, lon1, lat2, lon2, weight)
    else:
        return greatcircle_interpolate(lat1, lon1, lat2, lon2, weight)


def greatcircle_interpolate(lat1, lon1, lat2, lon2, weight):
    if (lat1 == lat2) and (lon1 == lon2):
        lat = lat1
        lon = lon1
    else:
        # transform to radians
        lat1 = lat1 * math.pi / 180
        lat2 = lat2 * math.pi / 180
        lon1 = lon1 * math.pi / 180
        lon2 = lon2 * math.pi / 180

        # initial bearing
        y = math.sin(lon2 - lon1) * math.cos(lat2)
        x = math.cos(lat1) * math.sin(lat2) - math.sin(lat1) * math.cos(lat2) * math.cos(lon2 - lon1)
        bearing = math.atan2(y, x)

        # angular distance from A to B
        d_tot = math.acos(math.sin(lat1) * math.sin(lat2) + math.cos(lat1) * math.cos(lat2) * math.cos(lon2 - lon1))
        d = d_tot * weight

        lat = math.asin(math.sin(lat1) * math.cos(d) + math.cos(lat1) * math.sin(d) * math.cos(bearing))
        lon = lon1 + math.atan2(
            math.sin(bearing) * math.sin(d) * math.cos(lat1), math.cos(d) - math.sin(lat1) * math.sin(lat)
        )

        # tansform to degrees
        lat = lat * 180 / math.pi
        lon = lon * 180 / math.pi
    return (lat, lon)


def mercator_projection(lat, lon):
    """Convert latitude and longitude to Mercator x, y coordinates."""
    R = 6378137  # Earth's radius in meters (WGS84)
    x = R * math.radians(lon)
    y = R * math.log(math.tan(math.pi / 4 + math.radians(lat) / 2))
    return x, y


def inverse_mercator_projection(x, y):
    """Convert Mercator x, y coordinates back to latitude and longitude."""
    R = 6378137
    lon = math.degrees(x / R)
    lat = math.degrees(2 * math.atan(math.exp(y / R)) - math.pi / 2)
    return lat, lon


def mercator_interpolate(lat1, lon1, lat2, lon2, fraction):
    """Find the point that lies 'fraction' along the straight line in Mercator projection."""
    x1, y1 = mercator_projection(lat1, lon1)
    x2, y2 = mercator_projection(lat2, lon2)

    x_interp = x1 + (x2 - x1) * fraction
    y_interp = y1 + (y2 - y1) * fraction

    (lat, lon) = inverse_mercator_projection(x_interp, y_interp)
    return (lat, lon)



import copy

import numpy as np


class FullProfiles:
    """
    FullProfiles is used to store the full profiles in a data structure, to avoid repeatedly reading
    in from file between fault situations.
    """

    def readProfileDataAndStore(
        self, griddata, filename, full_timerange, storagevalue_filling=None, storagevalue_time=None, timedelta=1.0
    ):
        self.stored_profiles = griddata._readProfileFromFile(filename, full_timerange)
        self.timedelta = timedelta
        if storagevalue_filling is not None:
            self.stored_storagevalue_time = griddata._readProfileFromFile(storagevalue_time, full_timerange)
            self.storagevalue_filling = griddata._readStoragevaluesFromFile(storagevalue_filling)

    def setProfileData(self, timerange, griddata):
        """timerange is here relative to full_timerange.
        If full_timerange does not start at zero this can give wrong results
        """
        profiles = self.stored_profiles.loc[timerange]
        profiles.index = range(len(timerange))
        griddata.profiles = profiles

        if "stored_storagevalue_time" in self.__dict__:
            storagevalue_time = self.stored_storagevalue_time.loc[timerange]
            storagevalue_time.index = range(len(timerange))
            griddata.storagevalue_time = storagevalue_time

        griddata.timerange = timerange
        griddata.timeDelta = self.timedelta

    def get_multiplier_demand(self):
        if "_multiplier_demand" in self.__dict__.keys():
            return self._multiplier_demand
        multiplier_demand = {}
        for demand_ref in self.stored_profiles.columns:
            if demand_ref == "index":
                print("Skipping index")
            elif demand_ref[:4] == "load":
                sum_demand = np.sum(self.stored_profiles[demand_ref])
                multiplier_demand[demand_ref] = sum_demand
        self._multiplier_demand = multiplier_demand
        return self._multiplier_demand


def get_gridmodel_failure(full_profiles, gridmodel_base, timerange, failed_generators, failed_branches):
    """Get a version of the gridmodel that takes into account the fault situation.

    timerange : relative to full model

    """

    gridmodel = copy.deepcopy(gridmodel_base)
    full_profiles.setProfileData(timerange, gridmodel)

    # Set failed generators to zero
    for gg in failed_generators:
        # if gg is an int, set that generator to zero
        if isinstance(gg, int):
            gridmodel.generator.at[gg, "pmax"] = 0.0
        elif isinstance(gg, tuple):
            # otherwise we've specified both the generator and the fraction that goes offline
            g_idx, g_down = gg
            gridmodel.generator.at[g_idx, "pmax"] *= 1 - g_down
        else:
            raise ValueError
    # Set failed branches to zero
    # for bb in failed_branches:
    #     gridmodel.branch.at[bb, 'capacity'] = 0
    # Remove failed branches
    to_keep = [ii for ii in range(gridmodel_base.numBranches()) if ii not in failed_branches]
    gridmodel.branch = copy.deepcopy(gridmodel_base.branch.iloc[to_keep])
    gridmodel.branch = gridmodel.branch.reset_index()

    return gridmodel



import copy

import numpy as np
import pyomo.opt

import powergama

"""
This file includes versions of LpProblem, Results and DataBase that are adjusted to running fault simulations.
"""


def extract_init(base_case, N_init, ii):
    storage_init = np.zeros(N_init)
    # Need to process the data from database into same format
    # Off by one (ii-1) since these are storage levels as stored as results for next step
    for _, rr in base_case[base_case.timestep == ii - 1].iterrows():
        storage_init[int(rr.indx)] = rr.storage
    return storage_init


class LpFaultProblem(powergama.LpProblem):
    """
    Modification of powergama LpProblem, to be run for non-sequential time-steps and
    import storage levels from previous runs
    """

    def __init__(self, grid, storage_base, flexload_base, timesteps=None, lossmethod=0, fault_duration=1):
        self._storage_base = storage_base
        self._flexload_base = flexload_base
        self._timesteps = timesteps
        self._fault_duration = fault_duration
        self._prev_ts = None
        self._dur_counter = 0
        self._counter = 0

        if lossmethod not in [0, 1]:
            raise ValueError("Not implemented for lossmethod 2 yet")
        super().__init__(grid, lossmethod)

    def _starting_timesteps_to_solve(self):
        if self._timesteps is None:
            numTimesteps = len(self._grid.timerange)
            return range(numTimesteps)
        else:
            return self._timesteps

    def _get_timesteps_to_solve(self, continue_from_last=False, results=None):
        if continue_from_last:
            raise ValueError("continue from last is not allowed with fault scenario simulation")
        starting_timesteps = self._starting_timesteps_to_solve()
        max_ts = self._grid.timerange[-1] + 1
        return [st + delay for st in starting_timesteps for delay in range(self._fault_duration) if st + delay < max_ts]

    def _get_fault_start(self, timestep):
        # dur_counter has already been incremented in _update_progress hence + 1
        return timestep - self._dur_counter + 1

    def reset_from_ts(self, timestep, storage_base, flexload_base):
        if timestep > 0:
            storage_init = extract_init(storage_base, len(self._grid.generator), timestep)
            flexload_init = extract_init(flexload_base, len(self._grid.consumer), timestep)

            storagecapacity = self._grid.generator["storage_cap"]
            self._storage = np.maximum(0, np.minimum(storagecapacity, storage_init))
            self._storage_flexload = flexload_init

    def _updateLpProblem(self, timestep):
        end_ts = self._grid.timerange[-1]
        if (self._prev_ts is None) or (self._dur_counter == self._fault_duration):
            reverting = True
        elif self._prev_ts == end_ts:  # last timestep reached end of possible simulation
            if not (timestep <= end_ts):
                raise Exception("timestep<=end_ts")
            reverting = True
        else:
            if not (self._prev_ts + 1 == timestep):
                raise Exception("self._prev_ts + 1 == timestep")
            reverting = False
        if reverting:
            self.reset_from_ts(timestep, self._storage_base, self._flexload_base)
            self._dur_counter = 0
        super()._updateLpProblem(timestep)

    def _update_progress(self, cur_step, num_steps):
        self._dur_counter += 1
        self._prev_ts = cur_step
        self._counter += 1
        super()._update_progress(self._counter - 1, num_steps)

    def _relax_and_retry(self, opt, warmstart, count, solve_args):
        print("Trying again without minimum power generation constraints.")
        old_pmin = copy.deepcopy(self._grid.generator["pmin"])
        N_gen = len(self._grid.generator)
        self._grid.generator["pmin"] = np.zeros(N_gen)
        old_p_gen_pmin = copy.deepcopy(self.p_gen_pmin)
        for i in range(N_gen):
            self.p_gen_pmin[i] = 0

        warmstart_now = warmstart and (count > 0)
        if warmstart_now:
            res = opt.solve(self, warmstart=warmstart_now, **solve_args)
        else:
            res = opt.solve(self, **solve_args)
        self.solver_res = res

        if (res.solver.status != pyomo.opt.SolverStatus.ok) or (
            res.solver.termination_condition == pyomo.opt.TerminationCondition.infeasible
        ):
            raise Exception("No solution after relaxing problem.")

        # Reset settings
        self._grid.generator["pmin"] = old_pmin
        for i in range(N_gen):
            self.p_gen_pmin[i] = old_p_gen_pmin[i]


class FaultDatabase(powergama.database.DatabaseBaseClass):
    def __init__(self, filename):
        super().__init__(filename)
        self.timestep_str = "timestep INT, fault_start INT"
        self.timestep_qs = "?,?"

    def timestep_tuple(self, timestep, fault_start):
        print(timestep, fault_start)
        return (timestep, fault_start)

    def extra_something(self):
        print("Bonus!")

    # TODO: might need updating?
    # def getTimerange(self):


class FaultResults(powergama.ResultsBaseClass):
    def _init_database(self, databasefile):
        self.db = FaultDatabase(databasefile)



"""
PowerGAMA Fault scenario simulation module
"""

from . import loadshedding_stats, specify_storage
from .failure_situation import FaultSpec, collect_res_failure_situation
from .full_profiles import FullProfiles
from .generate_failure_situations import create_fault_scenarios, run_fault_simulation
from .LpFaultProblem import LpFaultProblem



"""
Module containing the PowerGAMA Results class
"""

import csv
import math

import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import powergama.database as db
import powergama.GridData


class ResultsBaseClass(object):
    """
    Class for storing and analysing/presenting results from PowerGAMA
    Base version without plotting or results extraction

    Parameters
    ----------
    grid : GridData
        PowerGAMA GridData object
    databasefile : string
        name of sqlite3 file for storage of results
    replace : boolean
        whether to replace existing sqlite file (default=true).
        replace=false is useful to analyse previously
        generated results
    """

    def __init__(self, grid, databasefile, replace=True):
        """
        Create a PowerGAMA Results object

        """
        self.grid = grid
        self.timerange = grid.timerange
        self.storage_idx_generators = grid.getIdxGeneratorsWithStorage()
        self.pump_idx_generators = grid.getIdxGeneratorsWithPumping()
        self.flex_idx_consumers = grid.getIdxConsumersWithFlexibleLoad()
        self.idxConstrainedBranchCapacity = grid.getIdxBranchesWithFlowConstraints()

        self._init_database(databasefile)
        if replace:
            self.db.createTables(grid)
        else:
            # check that the length of the specified timerange matches the
            # database
            timerange_db = self.db.getTimerange()
            if not timerange_db:
                print(f"OBS: Empty database time range (no simulation run) = {timerange_db}")
            elif timerange_db != list(self.timerange):
                print(f"OBS: Database time range = [{timerange_db[0], timerange_db[-1]}]\n")
                # raise Exception("Database time range mismatch")

    def _init_database(self, databasefile):
        self.db = db.Database(databasefile)

    @classmethod
    def from_existing(cls, databasefile, timedelta):
        """Get PowerGAMA Result object from existing sqlite database file"""
        res_db = db.Database(databasefile)
        data_dict = res_db.get_grid_data()
        grid_data = powergama.GridData()
        grid_data.from_dict(data_dict, timedelta=timedelta)
        res = cls(grid_data, databasefile, replace=False)
        return res

    def addResultsFromTimestep(
        self,
        timestep,
        objective_function,
        generator_power,
        generator_pumped,
        branch_power,
        dcbranch_power,
        node_angle,
        sensitivity_branch_capacity,
        sensitivity_dcbranch_capacity,
        sensitivity_node_power,
        storage,
        inflow_spilled,
        loadshed_power,
        marginalprice,
        flexload_power,
        flexload_storage,
        flexload_storagevalue,
        branch_ac_losses=None,
        branch_dc_losses=None,
        fault_start=None,
    ):
        """Store results from optimal power flow for a new timestep

        timestep : int
            timestamp of results
        objective_function : float
            value of objective function
        generator_power : list
            generator output (list of same length and order as generators)
        generator_pumped : list
            position according to grid.getIdxGeneratorsWithPumping()
        branch_power : list
            AC branch power flow (list of same length and order as branches)
        dcbranch_power : list
            DC branch power flow (list of same length and order as dcbranches)
        node_angles : list
            voltage angles (list of same length and order as nodes)
        sensitivity_branch_capacity : list
            dual, position according to grid.getIdxBranchesWithFlowConstraints()
        sensitivity_dcbranch_capacity : list
            dual, capacity (list of same length and order as dcbranches)
        sensitivity_node_power : list
            dual, node demand(list of same length and order as nodes)
        storage : list
            position accordint to grid.getIdxGeneratorsWithStorage()
        inflow_spilled : list
            spileld power (list of same length and order as generators)
        loadshed_power : list
            same length and order as nodes
        marginalprice : list
            position according to grid.getIdxGeneratorsWithStorage()
        flexload_power : list
            position according to grid.getIdxConsumersWithFlexibleLoad()
        flexload_storage : list
            position according to grid.getIdxConsumersWithFlexibleLoad()
        flexload_storagevalue : list
            position according to grid.getIdxConsumersWithFlexibleLoad()
        branch_ac_losses : list
            ac branch losses
        branch_dc_losses : list
            dc branch losses
        fault_start  (int)
            extra identifier for when simulations are run for several timesteps
            from multiple starting timepoints
        """
        # Use zero if no branch power losses given:
        if branch_ac_losses is None:
            branch_ac_losses = [0] * len(branch_power)
        if branch_dc_losses is None:
            branch_dc_losses = [0] * len(dcbranch_power)
        # Store results in sqlite database on disk (to avoid memory problems)
        self.db.appendResults(
            timestep=timestep,
            objective_function=objective_function,
            generator_power=generator_power,
            generator_pumped=generator_pumped,
            branch_flow=branch_power,
            dcbranch_flow=dcbranch_power,
            node_angle=node_angle,
            sensitivity_branch_capacity=sensitivity_branch_capacity,
            sensitivity_dcbranch_capacity=sensitivity_dcbranch_capacity,
            sensitivity_node_power=sensitivity_node_power,
            storage=storage,
            inflow_spilled=inflow_spilled,
            loadshed_power=loadshed_power,
            marginalprice=marginalprice,
            flexload_power=flexload_power,
            flexload_storage=flexload_storage,
            flexload_storagevalue=flexload_storagevalue,
            idx_storagegen=self.storage_idx_generators,
            idx_branchsens=self.idxConstrainedBranchCapacity,
            idx_pumpgen=self.pump_idx_generators,
            idx_flexload=self.flex_idx_consumers,
            branch_ac_losses=branch_ac_losses,
            branch_dc_losses=branch_dc_losses,
            fault_start=fault_start,
        )

        """
        self.objectiveFunctionValue.append(objective_function)
        self.generatorOutput.append(generator_power)
        self.branchFlow.append(branch_power)
        self.nodeAngle.append(node_angle)
        self.sensitivityBranchCapacity.append(sensitivity_branch_capacity)
        self.sensitivityDcBranchCapacity.append(sensitivity_dcbranch_capacity)
        self.sensitivityNodePower.append(sensitivity_node_power)
        self.storage.append(storage)
        self.inflowSpilled.append(inflow_spilled)
        self.loadshed.append(loadshed_power)
        self.marginalprice.append(marginalprice)
        """
        # self.storageGeneratorsIdx.append(idx_generatorsWithStorage)


class Results(ResultsBaseClass):
    """
    Class for storing and analysing/presenting results from PowerGAMA

    Parameters
    ----------
    grid : GridData
        PowerGAMA GridData object
    databasefile : string
        name of sqlite3 file for storage of results
    replace : boolean
        whether to replace existing sqlite file (default=true).
        replace=false is useful to analyse previously
        generated results
    """

    def get_last_timestep_in_results(self):
        """Get last timestep for which results are stored."""
        time_last = None
        timerange = self.db.getTimerange()
        if timerange:
            # list is non-empty, get the last value:
            time_last = timerange[-1]
        return time_last

    def getAverageBranchFlows(self, timeMaxMin=None, branchtype="ac"):
        """
        Average flow on branches over a given time period

        Parameters
        ==========
        timeMaxMin : list (default = None)
            [min, max] - lower and upper time interval
        branchtype : string
            'ac' (default) or 'dc'

        Returns
        =======
        List with values for each branch:
        [flow from 1 to 2, flow from 2 to 1, average absolute flow]
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        # branchflow = self.db.getResultBranchFlowAll(timeMaxMin)
        if branchtype == "ac":
            ac = True
        elif branchtype == "dc":
            ac = False
        else:
            raise Exception('Branch type must be "ac" or "dc"')

        avgflow = self.db.getResultBranchFlowsMean(timeMaxMin, ac)
        # np.mean(branchflow,axis=1)
        return avgflow

    def getNodalPrices(self, node, timeMaxMin=None):
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        prices = self.db.getResultNodalPrice(node, timeMaxMin)
        # use asarray to convert None to nan
        prices = np.asarray(prices, dtype=float)
        return prices

    def getAverageNodalPrices(self, timeMaxMin=None):
        """
        Average nodal price over a given time period

        Parameters
        ----------
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval

        Returns
        =======
        1-dim Array of nodal prices (one per node)
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        avgprices = self.db.getResultNodalPricesMean(timeMaxMin)
        # use asarray to convert None to nan
        avgprices = np.asarray(avgprices, dtype=float)
        return avgprices

    def getAreaPrices(self, area, timeMaxMin=None):
        """
        Weighted average nodal price timeseries for given area
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        # area_nodes = [n._i for n in self.grid.node if n.area==area]
        loads = self.grid.getConsumersPerArea()[area]
        node_weight = [0] * len(self.grid.node["id"])
        for ld in loads:
            the_node = self.grid.consumer["node"][ld]
            the_load = self.grid.consumer["demand_avg"][ld]
            node_indx = self.grid.node["id"].tolist().index(the_node)
            node_weight[node_indx] += the_load

        sumWght = sum(node_weight)
        node_weight = [a / sumWght for a in node_weight]

        # print("Weights:")
        # print(node_weight)
        prices = self.db.getResultAreaPrices(node_weight, timeMaxMin)

        return prices

    def getAreaPricesAverage(self, areas=None, timeMaxMin=None):
        """
        Time average of weighted average nodal price per area
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        if areas is None:
            areas = self.grid.getAllAreas()

        avg_nodalprices = self.getAverageNodalPrices(timeMaxMin)
        all_loads = self.grid.getConsumersPerArea()
        avg_areaprice = {}

        for area in areas:
            nodes_in_area = [i for i, n in enumerate(self.grid.node.area) if n == area]
            node_weight = [0] * len(self.grid.node.id)
            if area in all_loads:
                loads = all_loads[area]
                for ld in loads:
                    the_node = self.grid.consumer.node[ld]
                    the_load = self.grid.consumer.demand_avg[ld]
                    node_indx = self.grid.node.id.tolist().index(the_node)
                    node_weight[node_indx] += the_load
                sumWght = sum(node_weight)
                node_weight = [a / sumWght for a in node_weight]

                prices = [node_weight[i] * avg_nodalprices[i] for i in nodes_in_area]
            else:
                # flat weight if there are no loads in area
                prices = [avg_nodalprices[i] for i in nodes_in_area]
            avg_areaprice[area] = sum(prices)

        return avg_areaprice

    def getLoadheddingInArea(self, area, timeMaxMin=None):
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        loadshed = self.db.getResultLoadheddingInArea(area, timeMaxMin)
        # use asarray to convert None to nan
        loadshed = np.asarray(loadshed, dtype=float)
        return loadshed

    def getLoadsheddingPerNode(self, timeMaxMin=None, average=False):
        """get loadshedding sum per node"""
        timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        loadshed_per_node = self.db.getResultLoadheddingSum(timeMaxMin, average=average)
        return loadshed_per_node

    def getLoadheddingSums(self, timeMaxMin=None, average=False):
        """get loadshedding sum per area"""
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        loadshed_per_node = self.db.getResultLoadheddingSum(timeMaxMin, average=average)
        areas = self.grid.node.area
        allareas = self.grid.getAllAreas()
        loadshed_sum = dict()
        for a in allareas:
            loadshed_sum[a] = sum([loadshed_per_node[i] for i in range(len(areas)) if areas[i] == a])

        # loadshed_sum = np.asarray(loadshed_sum,dtype=float)
        return loadshed_sum

    def getAverageEnergyBalance(self, timeMaxMin=None):
        """
        Average energy balance (generation minus demand) over a time period

        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval

        Returns
        =======
        1-dim Array of nodal prices (one per node)
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        branchflows = self.db.getResultBranchFlowsMean(timeMaxMin)
        if self.grid.numDcBranches() > 0:
            branchflowsDc = self.db.getResultBranchFlowsMean(timeMaxMin, ac=False)
        br_from = self.grid.branchFromNodeIdx()
        br_to = self.grid.branchToNodeIdx()
        dcbr_from = self.grid.dcBranchFromNodeIdx()
        dcbr_to = self.grid.dcBranchToNodeIdx()
        energybalance = []
        for n in range(len(self.grid.node["id"])):
            idx_from = [i for i, x in enumerate(br_from) if x == n]
            idx_to = [i for i, x in enumerate(br_to) if x == n]
            dc_idx_from = [i for i, x in enumerate(dcbr_from) if x == n]
            dc_idx_to = [i for i, x in enumerate(dcbr_to) if x == n]
            energybalance.append(
                sum([branchflows[0][i] - branchflows[1][i] for i in idx_from])
                - sum([branchflows[0][j] - branchflows[1][j] for j in idx_to])
                + sum([branchflowsDc[0][i] - branchflowsDc[1][i] for i in dc_idx_from])
                - sum([branchflowsDc[0][j] - branchflowsDc[1][j] for j in dc_idx_to])
            )

        # use asarray to convert None to nan
        energybalance = np.asarray(energybalance, dtype=float)
        return energybalance

    def getAverageBranchSensitivity(self, timeMaxMin=None, branchtype="ac"):
        """
        Average branch capacity sensitivity over a given time period

        Parameters
        ----------
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        branchtype : str
            ac or dc branch type

        Returns
        =======
        1-dim Array of sensitivities (one per branch)
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        avgsense = self.db.getResultBranchSensMean(timeMaxMin, branchtype)
        # use asarray to convert None to nan
        avgsense = np.asarray(avgsense, dtype=float)
        return avgsense

    def getAverageUtilisation(self, timeMaxMin=None, branchtype="ac"):
        """
        Average branch utilisation over a given time period

        Parameters
        ----------
        timeMaxMin :  (list) (default = None)
            [min, max] - lower and upper time interval
        branchtype : str
            ac or dc branch type

        Returns
        =======
        1-dim Array of branch utilisation (power flow/capacity)
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        if branchtype == "ac":
            cap = self.grid.branch.capacity
        elif branchtype == "dc":
            cap = self.grid.dcbranch.capacity
        avgflow = self.getAverageBranchFlows(timeMaxMin, branchtype)[2]
        utilisation = [avgflow[i] / cap.iloc[i] for i in range(len(cap))]
        utilisation = np.asarray(utilisation)
        return utilisation

    def getSystemCost(self, timeMaxMin=None):
        """
        Calculates system cost for energy produced by using generator fuel cost.

        Parameters
        ----------
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval

        Returns
        -------
        array of dictionary of cost of generation sorted per area
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        generation_per_gen = self.db.getResultGeneratorPowerSum(timeMaxMin)
        fuelcost_per_gen = self.grid.generator["fuelcost"]
        areas_per_gen = [
            self.grid.node["area"][self.grid.node["id"] == n].tolist()[0] for n in self.grid.generator["node"]
        ]

        allareas = self.grid.getAllAreas()
        generationcost = dict()
        for a in allareas:
            generationcost[a] = sum(
                [
                    generation_per_gen[i] * fuelcost_per_gen[i]
                    for i in range(len(areas_per_gen))
                    if areas_per_gen[i] == a
                ]
            )

        return generationcost

    def getGeneratorOutputSumPerArea(self, timeMaxMin=None):
        """
        Description
        Sums up generation per area.

        Parameters
        ----------
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval

        Returns
        =======
        array of dictionary of generation sorted per area
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        generation_per_gen = self.db.getResultGeneratorPowerSum(timeMaxMin)
        areas_per_gen = [self.grid.node.area[self.grid.node.id == n].tolist()[0] for n in self.grid.generator.node]

        allareas = self.grid.getAllAreas()
        generation = dict()
        for a in allareas:
            generation[a] = sum([generation_per_gen[i] for i in range(len(areas_per_gen)) if areas_per_gen[i] == a])

        return generation

    def getGeneratorSpilledSums(self, timeMaxMin=None):
        """Get sum of spilled inflow for all generators

        Parameters
        ----------
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        v = self.db.getResultGeneratorSpilledSums(timeMaxMin)
        return v

    def getGeneratorSpilled(self, generatorindx, timeMaxMin=None):
        """Get spilled inflow time series for given generator

        Parameters
        ----------
        generatorindx (int)
            index ofgenerator
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        v = self.db.getResultGeneratorSpilled(generatorindx, timeMaxMin)
        return v

    def getGeneratorStorageAll(self, timestep):
        """Get stored energy for all storage generators at given time

        Parameters
        ----------
        timestep : int
            timestep when storage is requested
        """
        v = self.db.getResultStorageFillingAll(timestep)

        return v

    def getGeneratorStorageValues(self, timestep):
        """Get value of stored energy for given time

        Parameters
        ----------
        timestep : int
            when to compute value

        Returns
        -------
        list of int
            Value of stored energy for all storage generators

        The method uses the storage value absolute level (basecost) per
        generator to compute total storage value
        """
        storage_energy = self.getGeneratorStorageAll(timestep)
        storage_values = self.grid.generator.storage_price
        indx_storage_generators = self.grid.getIdxGeneratorsWithStorage()
        storval = [storage_energy[i] * storage_values[v] for i, v in enumerate(indx_storage_generators)]
        return storval

    def _node2area(self, nodeName):
        """Returns the area of a spacified node"""
        # Is handy when you need to access more information about the node,
        # but only the node name is avaiable. (which is the case in the generator file)
        area = self.grid.node.loc[self.grid.node["id"] == nodeName, "area"].iloc[0]
        return area

    def _getAreaTypeProduction(self, area, generatorType, timeMaxMin):
        """
        Returns total production for specified area nd generator type
        """

        print("Looking for generators of type " + str(generatorType) + ", in " + str(area))
        print("Number of generator to run through: " + str(self.grid.generator.numGenerators()))
        totalProduction = 0

        for genNumber in range(0, self.grid.generator.numGenerators()):
            genNode = self.grid.generator.node[genNumber]
            genType = self.grid.generator.type[genNumber]
            genArea = self._node2area(genNode)
            # print str(genNumber) + ", " + genName + ", " + genNode + ", " + genType + ", " + genArea
            if (genType == generatorType) and (genArea == area):
                # print "\tGenerator is of right type and area. Adding production"
                genProd = sum(self.db.getResultGeneratorPower(genNumber, timeMaxMin))
                totalProduction += genProd
                # print "\tGenerator production = " + str(genProd)
        return totalProduction

    def getAllGeneratorProductionOBSOLETE(self, timeMaxMin=None):
        """Returns all production [MWh] for all generators"""
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        totGenNumbers = self.grid.generator.numGenerators()
        totalProduction = 0
        for genNumber in range(0, totGenNumbers):
            genProd = sum(self.db.getResultGeneratorPower(genNumber, timeMaxMin))
            print(str(genProd))
            totalProduction += genProd
            print("Progression: " + str(genNumber + 1) + " of " + str(totGenNumbers))
        return totalProduction

    def _productionOverview(self, areas, types, timeMaxMin, TimeUnitCorrectionFactor):
        """
        Returns a matrix with sum of generator production per area and type

        This function is manly used as the calculation part of the
        writeProductionOverview Contains just numbers (production[MWH] for
        each type(columns) and area(rows)), not headers
        """

        numAreas = len(areas)
        numTypes = len(types)
        resultMatrix = np.zeros((numAreas, numTypes))
        for areaIndex in range(0, numAreas):
            for typeIndex in range(0, numTypes):
                prod = self._getAreaTypeProduction(areas[areaIndex], types[typeIndex], timeMaxMin)
                print(
                    "Total produced " + types[typeIndex] + " energy for " + areas[areaIndex] + " equals: " + str(prod)
                )
                resultMatrix[areaIndex][typeIndex] = prod * TimeUnitCorrectionFactor
        return resultMatrix

    def writeProductionOverview(self, areas, types, filename=None, timeMaxMin=None, TimeUnitCorrectionFactor=1):
        """
        Export production overview to CSV file

        Write a .csv overview of the production[MWh] in timespan 'timeMaxMin'
        with the different areas and types as headers.
        The vectors 'areas' and 'types' becomes headers (column- and row
        headers), but the different elements
        of 'types' and 'areas' are also the key words in the search function
        'getAreaTypeProduction'.
        The vectors 'areas' and 'types' can be of any length.
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        corner = "Countries"
        numAreas = len(areas)
        numTypes = len(types)
        prodMat = self._productionOverview(areas, types, timeMaxMin, TimeUnitCorrectionFactor)
        if filename is not None:
            with open(filename, "wb") as f:
                writer = csv.writer(f)
                types.insert(0, corner)
                writer.writerow(types)
                for i in range(0, numAreas):
                    row = [areas[i]]
                    for j in range(0, numTypes):
                        row.append(str(prodMat[i][j]))
                    writer.writerow(row)
        else:
            title = ""
            for j in types:
                title = title + "\t" + j
            print("Area" + title)
            for i in range(0, numAreas):
                print(areas[i] + "\t%s" % "\t".join(map(str, prodMat[i])))

    def getAverageInterareaBranchFlow(self, filename=None, timeMaxMin=None):
        """Calculate average flow in each direction and total flow for
        inter-area branches.

        Parameters
        ----------
        filename : string, optional
            if a filename is given then the information is stored to file.
        timeMaxMin : list with two integer values, or None, optional
            time interval for the calculation [start,end]

        Returns
        -------
        List with values for each inter-area branch:
        [flow from 1 to 2, flow from 2 to 1, average absolute flow]
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        try:
            results = self.db.getAverageInterareaBranchFlow(timeMaxMin)
        except Exception as err:
            print("Error occured. Maybe because you are using sqlite<3.7")
            raise (err)

        if filename is not None:
            headers = ("branch", "fromArea", "toArea", "average negative flow", "average positive flow", "average flow")
            with open(filename, "wb") as f:
                writer = csv.writer(f)
                writer.writerow(headers)
                for row in results:
                    writer.writerow(row)

        return results

    def getAverageImportExport(self, area, timeMaxMin=None):
        """Return average import and export for a specified area"""

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        ia = self.getAverageInterareaBranchFlow(timeMaxMin=timeMaxMin)

        # export: A->B pos flow + A<-B neg flow
        sum_export = sum([b[4] for b in ia if b[1] == area]) - sum([b[3] for b in ia if b[2] == area])
        # import: A->B neg flow + A<-B pos flow
        sum_import = -sum([b[3] for b in ia if b[2] == area]) + sum([b[4] for b in ia if b[2] == area])
        return dict(exp=sum_export, imp=sum_import)

    def getEnergyBalanceInArea(
        self,
        area,
        spillageGen,
        resolution="H",
        fileName=None,
        timeMaxMin=None,
        start_date="2014-01-01",
    ):
        """
        Print time series of energy balance in an area, including
        production, spillage, load shedding, storage, pump consumption
        and imports

        Parameters
        ----------
        area : string
            area code
        spillageGen : list
            generator types for which to show spillage (renewables)
        resolution : string
            resolution of output, see pandas:resample
        fileName : string (default=None)
            name of file to export results
        timeMaxMin : list
            time range to consider
        start_date : date string
            date when time series start

        """
        # Eirik/Arne

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        # data resolution in whole seconds (usually, timeDelta=1.0)
        resolutionS = int(self.grid.timeDelta * 3600)

        prod = pd.DataFrame()
        genTypes = self.grid.getAllGeneratorTypes()
        generators = self.grid.getGeneratorsPerAreaAndType()[area]
        pumpIdx = self.grid.getGeneratorsWithPumpByArea()
        if len(pumpIdx) > 0:
            pumpIdx = pumpIdx[area]
        storageGen = self.grid.getIdxGeneratorsWithStorage()
        areaGen = [item for sublist in list(generators.values()) for item in sublist]
        matches = [x for x in areaGen if x in storageGen]
        for gt in genTypes:
            if gt in generators:
                prod[gt] = self.db.getResultGeneratorPower(generators[gt], timeMaxMin)
                if gt in spillageGen:
                    prod[gt + " spilled"] = self.db.getResultGeneratorSpilled(generators[gt], timeMaxMin)
        prod["load shedding"] = self.getLoadheddingInArea(area, timeMaxMin)
        storage = self.db.getResultStorageFillingMultiple(matches, timeMaxMin, capacity=False)
        if storage:
            prod["storage"] = storage
        if len(pumpIdx) > 0:
            prod["pumped"] = self.db.getResultPumpPowerMultiple(pumpIdx, timeMaxMin, negative=True)
        prod["net import"] = self.getNetImport(area, timeMaxMin)
        prod.index = pd.date_range(start_date, periods=timeMaxMin[-1] - timeMaxMin[0], freq="{}s".format(resolutionS))
        if resolution != "H":
            prod = prod.resample(resolution, how="sum")
        if fileName:
            prod.to_csv(fileName)
        else:
            return prod

    def getStorageFillingInAreas(self, areas, generator_type, relative_storage=True, timeMaxMin=None):
        """
        Gets time-series with aggregated storage filling for specified area(s)
        for a specific generator type.

        Parameters
        ----------
        areas : list
            list of area codes (e.g. ['DE','FR'])
        generator_type : string
            generator type string (e.g. 'hydro')
        relative_storage : boolean
            show relative (True) or absolute (False) storage
        timeMaxMin : list
            time range to consider (e.g. [0,8760])
        """
        # Eirik/Arne

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        storageGen = self.grid.getIdxGeneratorsWithStorage()
        storageTypes = self.grid.generator.type
        nodeNames = self.grid.generator.node
        nodeAreas = self.grid.node.area
        storCapacities = self.grid.generator.storage_cap
        generators = []
        capacity = 0
        for gen in storageGen:
            area = nodeAreas[self.grid.node.id.tolist().index(nodeNames[gen])]
            if area in areas and storageTypes[gen] == generator_type:
                generators.append(gen)
                if relative_storage:
                    capacity += storCapacities[gen]
            filling = self.db.getResultStorageFillingMultiple(generators, timeMaxMin, capacity)
        return filling

    def getNetImport(self, area, timeMaxMin=None):
        """Return time series for net import for a specified area"""
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        # find the associated branches
        br = self.grid.getInterAreaBranches(area_to=area, acdc="ac")
        br_p = br["branches_pos"]
        br_n = br["branches_neg"]
        dcbr = self.grid.getInterAreaBranches(area_to=area, acdc="dc")
        dcbr_p = dcbr["branches_pos"]
        dcbr_n = dcbr["branches_neg"]

        # AC branches
        ie = self.db.getBranchesSumFlow(branches_pos=br_p, branches_neg=br_n, timeMaxMin=timeMaxMin, acdc="ac")
        # DC branches
        dcie = self.db.getBranchesSumFlow(branches_pos=dcbr_p, branches_neg=dcbr_n, timeMaxMin=timeMaxMin, acdc="dc")

        if ie["pos"] and ie["neg"]:
            res_ac = [a - b for a, b in zip(ie["pos"], ie["neg"])]
        elif ie["pos"]:
            res_ac = ie["pos"]
        elif ie["neg"]:
            res_ac = [-a for a in ie["neg"]]
        else:
            res_ac = [0] * (timeMaxMin[-1] - timeMaxMin[0])

        if dcie["pos"] and dcie["neg"]:
            res_dc = [a - b for a, b in zip(dcie["pos"], dcie["neg"])]
        elif dcie["pos"]:
            res_dc = dcie["pos"]
        elif dcie["neg"]:
            res_dc = [-a for a in dcie["neg"]]
        else:
            res_dc = [0] * (timeMaxMin[-1] - timeMaxMin[0])

        res = [a + b for a, b in zip(res_ac, res_dc)]
        return res

    def getImportExport(self, areas=None, timeMaxMin=None, acdc=["ac", "dc"]):
        """Return time series for import and export for a specified area"""
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        if areas is None:
            areas = self.grid.getAllAreas()

        df_importexport = pd.DataFrame(index=areas, columns=["import", "export"])
        for area in areas:
            print(area, end=",")
            # find the associated branches (pos = into area)
            # br = self.grid.getInterAreaBranches(area_to=area,acdc='ac')
            # br_p = br['branches_pos']
            # br_n = br['branches_neg']
            # dcbr = self.grid.getInterAreaBranches(area_to=area,acdc='dc')
            # dcbr_p = dcbr['branches_pos']
            # dcbr_n = dcbr['branches_neg']
            flow_in = 0
            flow_out = 0
            for acdc_type in acdc:
                br = self.grid.getInterAreaBranches(area_to=area, acdc=acdc_type)
                br_pos = self.db.getResultBranches(timeMaxMin, br_indx=br["branches_pos"], acdc=acdc_type)
                br_neg = self.db.getResultBranches(timeMaxMin, br_indx=br["branches_neg"], acdc=acdc_type)
                if br_pos.shape[0] > 0:
                    flow_in += br_pos[br_pos["flow"] > 0]["flow"].sum()
                    flow_out -= br_pos[br_pos["flow"] < 0]["flow"].sum()
                if br_neg.shape[0] > 0:
                    flow_in -= br_neg[br_neg["flow"] < 0]["flow"].sum()
                    flow_out += br_neg[br_neg["flow"] > 0]["flow"].sum()

            # ie =  self.db.getBranchesSumFlow(branches_pos=br_p,branches_neg=br_n,
            #                                 timeMaxMin=timeMaxMin,
            #                                 acdc='ac')
            # DC branches

            #            dcie =  self.db.getBranchesSumFlow(branches_pos=dcbr_p,
            #                                                 branches_neg=dcbr_n,
            #                                                 timeMaxMin=timeMaxMin,
            #                                                 acdc='dc')
            #            import_a = (sum(v for v in ie['pos'] if v>=0)
            #                         +sum(-v for v in ie['neg'] if v<0)
            #                         #+sum(v for v in dcie['pos'] if v>=0)
            #                         #+sum(-v for v in dcie['neg'] if v<0)
            #                         )
            #            export_a = (sum(-v for v in ie['pos'] if v<0)
            #                         +sum(v for v in ie['neg'] if v>=0)
            #                         #+sum(-v for v in dcie['pos'] if v<0)
            #                         #+sum(v for v in dcie['neg'] if v>=0)
            #                         )
            df_importexport.loc[area, "import"] = flow_in
            df_importexport.loc[area, "export"] = flow_out
        print()
        return df_importexport

    def getDemandPerArea(self, area, timeMaxMin=None):
        """Returns demand timeseries for given area, as dictionary with
        fields "fixed", "flex", and "sum"

        Parameters
        ----------
        area (string)
            area to get demand for
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        consumer = self.grid.consumer

        dem = [0] * len(self.timerange)
        flexdemand = [0] * len(self.timerange)
        consumers = self.grid.getConsumersPerArea()[area]
        for i in consumers:
            ref_profile = consumer.demand_ref[i]
            # accumulate demand for all consumers in this area:
            dem = [
                dem[t - self.timerange[0]]
                + consumer.demand_avg[i]
                * (1 - consumer.flex_fraction[i])
                * self.grid.profiles[ref_profile][t - self.timerange[0]]
                for t in timerange
            ]
            flexdemand_i = self.db.getResultFlexloadPower(i, timeMaxMin)
            if len(flexdemand_i) > 0:
                flexdemand = [sum(x) for x in zip(flexdemand, flexdemand_i)]
        sumdemand = [sum(x) for x in zip(dem, flexdemand)]

        return {"fixed": dem, "flex": flexdemand, "sum": sumdemand}

    def plotNodalPrice(self, nodeIndx, timeMaxMin=None, showTitle=True):
        """Show nodal price in single node

        Parameters
        ----------
        nodeIndx (int)
            index of node to plot from
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """

        # TODO allow for input to be multiple nodes
        # TODO plot storage price for storage in the same node?
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        nodalprice = self.db.getResultNodalPrice(nodeIndx, timeMaxMin)
        plt.figure()
        plt.plot(timerange, nodalprice)
        if showTitle:
            plt.title("Nodal price for node %d" % (nodeIndx))
        return

    def plotAreaPrice(self, areas, timeMaxMin=None, showTitle=True):
        """Show area price(s)

        Parameters
        ----------
        areas (list)
            list of areas to show
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        plt.figure()
        for a in areas:
            areaprice = self.getAreaPrices(a, timeMaxMin)
            plt.plot(timerange, areaprice, label=a)
            if showTitle:
                plt.title("Area price")

        plt.legend()
        return

    def plotStorageFilling(self, generatorIndx, timeMaxMin=None, showTitle=True):
        """Show storage filling level (MWh) for generators with storage

        Parameters
        ----------
        generatorIndx (int)
            index of generator to plot from
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        if generatorIndx in self.storage_idx_generators:
            storagefilling = self.db.getResultStorageFilling(generatorIndx, timeMaxMin)
            plt.figure()
            plt.plot(timerange, storagefilling)
            if showTitle:
                plt.title("Storage filling level for generator %d" % (generatorIndx))
        else:
            print("These are the generators with storage:")
            print(self.storage_idx_generators)
        return

    def plotGeneratorOutput(self, generator_index, timeMaxMin=None, relativestorage=True, showTitle=True):
        """Show output of a generator

        Parameters
        ----------
        generator_index (int)
            index of generator for which to make the plot
        timeMaxMin [int,int] (default=None)
            time interval for the plot [start,end]
        relativestorage (default=True)
            use filling fraction as y axis label for storage
        """
        # TODO allow for input to be multiple generators
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        generatoroutput = self.db.getResultGeneratorPower(generator_index, timeMaxMin)
        # powerinflow = db.getInflow(timeMaxMin,storageindx)

        ax1.plot(timerange, generatoroutput, "-r", label="output")
        ax1.set_ylim(ymin=0)

        # Power inflow (if generator has nonzero inflow factor)
        if self.grid.generator["inflow_fac"][generator_index] > 0:
            profile = self.grid.generator["inflow_ref"][generator_index]
            ax1.plot(
                timerange,
                [
                    self.grid.profiles[profile][t - self.timerange[0]]
                    * self.grid.generator["inflow_fac"][generator_index]
                    * self.grid.generator["pmax"][generator_index]
                    for t in timerange
                ],
                "-.b",
                label="inflow",
            )

        # Power pumped (if generator has nonzero pumping capacity)
        if self.grid.generator["pump_cap"][generator_index] > 0:
            pump_output = self.db.getResultPumpPower(generator_index, timeMaxMin)
            ax1.plot(timerange, pump_output, ":c", label="pumping")

        # Storage filling level (if generator has storage)
        ax2 = None
        if generator_index in self.storage_idx_generators:
            storagefilling = self.db.getResultStorageFilling(generator_index, timeMaxMin)
            if relativestorage:
                cap = self.grid.generator["storage_cap"][generator_index]
                storagefilling = [x / cap for x in storagefilling]
            ax2 = plt.twinx()  # separate y axis
            ax2.plot(timerange, storagefilling, "--g", label="storage")
            ax2.legend(loc="lower right", bbox_to_anchor=(1, 1), borderaxespad=0.0, frameon=False)
            ax2.set_ylim(ymin=0)

        ax1.legend(loc="lower left", bbox_to_anchor=(0, 1), borderaxespad=0.0, frameon=False)
        if ax2 is not None:
            # TODO: Add comment - What whas the point of this??
            # ax2.add_artist(lgd)
            # ax1.legend=None
            pass
        nodeidx = self.grid.node["id"].tolist().index(self.grid.generator["node"][generator_index])
        if showTitle:
            plt.title(
                "Generator %d (%s) at node %d (%s)"
                % (
                    generator_index,
                    self.grid.generator["type"][generator_index],
                    nodeidx,
                    self.grid.generator["node"][generator_index],
                )
            )
        return

    def plotDemandAtLoad(self, consumer_index, timeMaxMin=None, relativestorage=True, showTitle=True):
        """Make a time-series plot of consumption of a specified load

        Parameters
        ----------
        consumer_index (int)
            index of consumer for which to make the plot
        timeMaxMin [int,int] (default=None)
            time interval for the plot [start,end]
        relativestorage (default=True)
            use filling fraction as y axis label for storage
        """
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        # Fixed load
        profile = self.grid.consumer["demand_ref"][consumer_index]
        ax1.plot(
            timerange,
            [
                self.grid.profiles[profile][t - self.timerange[0]]
                * self.grid.consumer["demand_avg"][consumer_index]
                * (1 - self.grid.consumer["flex_fraction"][consumer_index])
                for t in timerange
            ],
            "-r",
            label="fixed load",
        )

        # Flexible load  (if consumer has nonzero flexible load)
        ax2 = None
        if self.grid.consumer.flex_fraction[consumer_index] > 0:
            flexload_power = self.db.getResultFlexloadPower(consumer_index, timeMaxMin)
            ax1.plot(timerange, flexload_power, "-.b", label="flexible load")

            # Storage filling level
            storagefilling = self.db.getResultFlexloadStorageFilling(consumer_index, timeMaxMin)
            if relativestorage:
                cap = self.grid.getFlexibleLoadStorageCapacity(consumer_index)
                storagefilling = [x / cap for x in storagefilling]
            ax2 = plt.twinx()  # separate y axis
            ax2.plot(timerange, storagefilling, "--g", label="storage")
            ax2.legend(loc="lower right", bbox_to_anchor=(1, 1), borderaxespad=0.0, frameon=False)

        ax1.legend(loc="lower left", bbox_to_anchor=(0, 1), borderaxespad=0.0, frameon=False)
        # if ax2 is not None:
        #    ax2.add_artist(lgd)
        #    ax1.legend=None
        nodeidx = self.grid.node["id"].tolist().index(self.grid.consumer.node[consumer_index])
        if showTitle:
            plt.title(
                "Consumer %d at node %d (%s)" % (consumer_index, nodeidx, self.grid.consumer.node[consumer_index])
            )
        return

    def plotStoragePerArea(self, area, absolute=False, timeMaxMin=None, showTitle=True):
        """Show generation storage accumulated per area

        Parameters
        ----------
        area (str)
        absolute (bool)(default=False)
            plot storage value in absolute or relative to maximum
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval"""

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        plt.figure()
        generators = self.grid.getGeneratorsPerAreaAndType()
        cap = self.grid.generator["storage_cap"]
        for gentype in generators[area].keys():
            idxGen = generators[area][gentype]
            idx_storage = [[i, v] for i, v in enumerate(self.storage_idx_generators) if v in idxGen]
            # idx_storage is now a list of index pairs for generators with
            #    storage in the given area
            # the first value is index in generator list
            # the second value is index in storage list (not used)

            if len(idx_storage) > 0:
                mystor = [
                    sum(
                        [
                            sum(self.db.getResultStorageFilling(idx_storage[i][1], [t, t + 1]))
                            for i in range(len(idx_storage))
                        ]
                    )
                    for t in timerange
                ]
                mycap = sum([cap[idx_storage[i][1]] for i in range(len(idx_storage))])

                if absolute:
                    sumStorAreaType = mystor
                else:
                    sumStorAreaType = [mystor[i] / mycap for i in range(len(mystor))]
                plt.plot(timerange, sumStorAreaType, label=gentype)

        # plt.legend(generators[area].keys() , loc="upper right")
        plt.legend(loc="upper right")
        if showTitle:
            plt.title("Total storage level in %s" % (area))

        return

    def plotGenerationPerArea(
        self, area, timeMaxMin=None, fill=True, reversed_order=False, net_import=True, loadshed=True, showTitle=True
    ):
        """Show generation per area

        Parameters
        ----------
        area (str)
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        fill (Boolean) - whether use filled plot
        reversed_order - whether to reverse order of generator types
        net_import - whether to include net import in graph
        loadshed - whether to include unmet demand
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])
        fillfrom = [0] * len(timerange)
        count = 0
        plt.figure()
        ax = plt.subplot(111)
        generators = self.grid.getGeneratorsPerAreaAndType()
        # gentypes_ordered = self._gentypes_ordered_by_fuelcost()
        gentypes_ordered = self.grid.getAllGeneratorTypes(sort="fuelcost")
        if reversed_order:
            gentypes_ordered.reverse()
        numCurves = len(gentypes_ordered) + 1
        # colours = cm.gist_rainbow(np.linspace(0, 1, numCurves))
        colours = cm.tab20([i % 20 for i in range(numCurves)])
        for gentype in gentypes_ordered:
            if gentype in generators[area]:
                idxGen = generators[area][gentype]
                sumGenAreaType = self.db.getResultGeneratorPower(idxGen, timeMaxMin)
                if fill:
                    aggregated = [x + y for x, y in zip(sumGenAreaType, fillfrom)]
                    ax.fill_between(timerange, y1=aggregated, y2=fillfrom, facecolor=colours[count])
                    # could add edgecolor="lightgrey"
                    # add this plot to get the legend right
                    ax.plot([], [], color=colours[count], linewidth=10, label=gentype)
                    fillfrom = aggregated
                else:
                    ax.plot(timerange, sumGenAreaType, color=colours[count], label=gentype)
            else:
                # in order to get the legend right
                ax.plot([], [], color=colours[count], label=gentype)
            count = count + 1
        if net_import:
            netimport = self.getNetImport(area, timeMaxMin)
            agg = [x + y for x, y in zip(netimport, fillfrom)]
            ax.plot(timerange, agg, linestyle=":", linewidth=2, color="black", label="net import")
        if loadshed:
            loadshed = self.getLoadheddingInArea(area, timeMaxMin)
            label = "Load shed"
            col = "dimgray"
            if fill:
                aggregated = [x + y for x, y in zip(loadshed, fillfrom)]
                # ax.fill_between(timerange,y1=aggregated,y2=fillfrom,
                #                facecolor=col)
                # ax.plot([],[],color=col,linewidth=10,
                #        label=label)
                ax.plot(timerange, aggregated, linestyle="--", color=col, label=label)
                fillfrom = aggregated
            else:
                ax.plot(timerange, loadshed, linestyle="--", color=col, label=label)

        # plt.legend()
        handles, labels = ax.get_legend_handles_labels()
        handles.reverse()
        labels.reverse()
        plt.legend(handles, labels, loc=2, bbox_to_anchor=(1.05, 1), borderaxespad=0.0)

        if fill:
            plt.ylim(ymin=0)

        if showTitle:
            plt.title("Generation in %s" % (area))
        return

    def plotDemandPerArea(self, areas, timeMaxMin=None, showTitle=True):
        """Show demand in area(s)

        Parameters
        ----------
        areas (list?)
            list of areas to be plotted
        timeMaxMin (list) (default = None)
            [min, max] - lower and upper time interval
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        plt.figure()
        consumer = self.grid.consumer
        if type(areas) is str:
            areas = [areas]
        for co in areas:
            dem = [0] * len(self.timerange)
            flexdemand = [0] * len(self.timerange)
            consumers = self.grid.getConsumersPerArea()[co]
            for i in consumers:
                ref_profile = consumer["demand_ref"][i]
                # accumulate demand for all consumers in this area:
                dem = [
                    dem[t - self.timerange[0]]
                    + consumer["demand_avg"][i]
                    * (1 - consumer["flex_fraction"][i])
                    * self.grid.profiles[ref_profile][t - self.timerange[0]]
                    for t in timerange
                ]
                flexdemand_i = self.db.getResultFlexloadPower(i, timeMaxMin)
                if len(flexdemand_i) > 0:
                    flexdemand = [sum(x) for x in zip(flexdemand, flexdemand_i)]
            sumdemand = [sum(x) for x in zip(dem, flexdemand)]
            (p,) = plt.plot(timerange, sumdemand, label=co)
            # Fixed demand in dotted lines
            plt.plot(timerange, dem, "--", color=p.get_color())

        plt.legend(loc="upper right")
        if showTitle:
            plt.title("Power demand")
        return

    def plotStorageValues(self, genindx, timeMaxMin=None, showTitle=True):
        """Plot storage values (marginal prices) for generators with storage

        Parameters
        ----------
        genindx (int)
            index of generator for which to make the plot
        timeMaxMin [int,int] (default=None)
            time interval for the plot [start,end]
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        if genindx in self.storage_idx_generators:
            nodeidx = self.grid.node["id"].tolist().index(self.grid.generator.node[genindx])
            storagevalue = self.db.getResultStorageValue(genindx, timeMaxMin)
            nodalprice = self.db.getResultNodalPrice(nodeidx, timeMaxMin)
            pumpprice = [x - self.grid.generator.pump_deadband[genindx] for x in storagevalue]
            plt.figure()
            (p,) = plt.plot(timerange, storagevalue, "-b", label="storage value")
            if genindx in self.pump_idx_generators:
                pumpprice = [x - self.grid.generator.pump_deadband[genindx] for x in storagevalue]
                plt.plot(timerange, pumpprice, ":", color=p.get_color(), label="pump threshold")
            plt.plot(timerange, nodalprice, "--r", label="nodal price")
            plt.legend()
            plt.legend(loc="lower left", bbox_to_anchor=(0, 1), borderaxespad=0.0, ncol=4, frameon=False)

            if showTitle:
                plt.title(
                    "Storage value  for generator %d (%s) in %s"
                    % (genindx, self.grid.generator.type[genindx], self.grid.generator.node[genindx])
                )
        else:
            print("These are the generators with storage:")
            print(self.storage_idx_generators)
        return

    def plotFlexibleLoadStorageValues(self, consumerindx, timeMaxMin=None, showTitle=True):
        """Plot storage valuesfor flexible loads

        Parameters
        ----------
        consumerindx : int
            index of consumer for which to make the plot
        timeMaxMin : list, [int,int]
            time interval for the plot [start,end], or None for entire range
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]
        timerange = range(timeMaxMin[0], timeMaxMin[-1])

        if consumerindx in self.flex_idx_consumers:
            nodeidx = self.grid.node.id.tolist().index(self.grid.consumer.node[consumerindx])
            storagevalue = self.db.getResultFlexloadStorageValue(consumerindx, timeMaxMin)
            nodalprice = self.db.getResultNodalPrice(nodeidx, timeMaxMin)
            plt.figure()
            plt.plot(timerange, storagevalue)
            plt.plot(timerange, nodalprice)
            plt.legend(["storage value", "nodal price"])
            if showTitle:
                plt.title(
                    "Storage value  for consumer %d at %s" % (consumerindx, self.grid.consumer.node[consumerindx])
                )
        else:
            print("These are the consumers with flexible load:")
            print(self.flex_idx_consumers)
        return

    def plotMapGrid(
        self,
        nodetype=None,
        branchtype=None,
        dcbranchtype=None,
        show_node_labels=False,
        branch_style="c",
        latlon=None,
        timeMaxMin=None,
        dotsize=40,
        filter_node=None,
        filter_branch=None,
        draw_par_mer=False,
        showTitle=True,
        colors=True,
    ):
        """
        Plot results to map

        Parameters
        ----------
        nodetype : string
            "", "area", "nodalprice", "energybalance", "loadshedding"
        branchtype : string
            "", "capacity", "area", "utilisation", "flow", "sensitivity"
        dcbranchtype : string
            "", "capacity"
        show_node_labels : boolean
            whether to show node names (true/false)
        branch_style : string or list of strings (optional)
            How branch capacity and flow should be visualised.
            "c" = colour, "t" = thickness. The two options may be combined.
        dotsize : integer (optional)
            set dot size for each plotted node
        latlon: list of four floats (optional)
            map area [lat_min, lon_min, lat_max, lon_max]
        filter_node : list of two floats (optional)
            [min,max] - lower and upper cutoff for node value
        filter_branch : list of two floats
            [min,max] - lower and upper cutoff for branch value
        draw_par_mer : boolean
            whether to draw parallels and meridians on map
        showTitle : boolean
        colors : boolean
            Whether to use colours or not
        """

        # basemap is only used here, so to allow using powergama without
        # basemap installed, it is best to put import statement here.
        from mpl_toolkits.basemap import Basemap

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        plt.figure()
        data = self.grid
        # res = self

        if latlon is None:
            lat_max = max(data.node.lat) + 1
            lat_min = min(data.node.lat) - 1
            lon_max = max(data.node.lon) + 1
            lon_min = min(data.node.lon) - 1
        else:
            lat_min = latlon[0]
            lon_min = latlon[1]
            lat_max = latlon[2]
            lon_max = latlon[3]

        # Use the average latitude as latitude of true scale
        lat_truescale = np.mean(data.node.lat)

        m = Basemap(
            resolution="l",
            projection="merc",
            lat_ts=lat_truescale,
            llcrnrlon=lon_min,
            llcrnrlat=lat_min,
            urcrnrlon=lon_max,
            urcrnrlat=lat_max,
            anchor="W",
        )

        # Draw coastlines, meridians and parallels.
        m.drawcoastlines()
        # m.drawcountries(zorder=0)

        if colors:
            m.fillcontinents(color="coral", lake_color="aqua", zorder=0)
            m.drawmapboundary(fill_color="aqua")
        else:
            m.fillcontinents(zorder=0)
            m.drawmapboundary()

        if draw_par_mer:
            m.drawparallels(
                np.arange(_myround(lat_min, 10, "floor"), _myround(lat_max, 10, "ceil"), 10), labels=[1, 1, 0, 0]
            )

            m.drawmeridians(
                np.arange(_myround(lon_min, 10, "floor"), _myround(lon_max, 10, "ceil"), 10), labels=[0, 0, 0, 1]
            )

        # AC Branches
        num_branches = data.branch.shape[0]

        lwidths = [2] * num_branches

        # default values:
        branch_value = np.asarray([0.5] * num_branches)
        branch_colormap = cm.gray
        branch_plot_colorbar = True

        if branchtype == "area":
            areas = data.node.area
            allareas = data.getAllAreas()
            branch_value = [-1] * num_branches
            nodes_from = data.branchFromNodeIdx()
            nodes_to = data.branchToNodeIdx()
            for i in range(num_branches):
                # node_indx_from = data.node.name.index(data.branch.node_from[i])
                # node_indx_to = data.node.name.index(data.branch.node_to[i])
                area_from = areas[nodes_from[i]]
                area_to = areas[nodes_to[i]]
                if area_from == area_to:
                    branch_value[i] = allareas.index(area_from)
                branch_value = np.asarray(branch_value)
            #            branch_colormap = plt.get_cmap('hsv')
            branch_colormap = cm.prism
            branch_colormap.set_under("k")
            filter_branch = [0, len(allareas)]
            branch_label = "Branch area"
            branch_plot_colorbar = False
        elif branchtype == "utilisation":
            utilisation = self.getAverageUtilisation(timeMaxMin)
            branch_value = utilisation
            branch_colormap = plt.get_cmap("hot")
            branch_label = "Branch utilisation"
        elif branchtype == "capacity":
            cap = data.branch.capacity
            branch_plot_colorbar = False
            branch_label = "Branch capacity"
            if "c" in branch_style:
                branch_value = np.asarray(cap)
                branch_colormap = plt.get_cmap("hot")
                branch_plot_colorbar = True
                if filter_branch is None:
                    # need an upper limit to avoid crash due to inf capacity
                    maxcap = np.nanmax(branch_value)
                    filter_branch = [0, np.round(maxcap, -2) + 100]
            if "t" in branch_style:
                avgcap = np.mean(cap)
                if avgcap == 0:
                    lwidths = [0 for f in cap]
                else:
                    lwidths = [2 * f / avgcap for f in cap]
        elif branchtype == "flow":
            avgflow = self.getAverageBranchFlows(timeMaxMin)[2]
            if "c" in branch_style:
                branch_value = np.asarray(avgflow)
                branch_colormap = plt.get_cmap("hot")
            else:
                branch_plot_colorbar = False
            if "t" in branch_style:
                avgavgflow = np.mean(avgflow)
                lwidths = [2 * f / avgavgflow for f in avgflow]
            branch_label = "Branch flow"
            # branch_value = np.asarray(avgflow)
            # branch_colormap = plt.get_cmap('hot')
        elif branchtype == "sensitivity":
            branch_value = np.zeros(num_branches)
            avgsense = self.getAverageBranchSensitivity(timeMaxMin)
            branch_value[self.idxConstrainedBranchCapacity] = -avgsense
            branch_colormap = plt.get_cmap("hot")
            branch_label = "Branch sensitivity"
            # These sensitivities are mostly negative
            # (reduced cost by increasing branch capacity)
            # minsense = np.nanmin(avgsense)
            # maxsense = np.nanmax(avgsense)
        else:
            branch_value = np.asarray([0.5] * num_branches)
            branch_colormap = cm.gray
            branch_plot_colorbar = False

        idx_from = data.branchFromNodeIdx()
        idx_to = data.branchToNodeIdx()
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)

        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        # ls = [[(x1[i],y1[i]),(x2[i],y2[i])] for i in range(num_branches)]
        ax = plt.axes()
        if not lwidths:
            print("No new data")
        else:
            line_segments_ac = mpl.collections.LineCollection(ls, linewidths=lwidths, cmap=branch_colormap)

            if filter_branch is not None:
                line_segments_ac.set_clim(filter_branch)
            line_segments_ac.set_array(branch_value)
            ax.add_collection(line_segments_ac)

        # DC Branches
        lwidths = [2] * 1
        # lwidths = [2] * num_dcbranches
        if dcbranchtype is not None:
            branch_plot_colorbar = False
            branch_value = np.asarray([0.1] * num_branches)
            branch_colormap = cm.winter

        if dcbranchtype == "flow":
            avgTot = self.getAverageBranchFlows(timeMaxMin, False)[2]
            if "c" in branch_style:
                branch_value = np.asarray(avgTot)
                branch_colormap = plt.get_cmap("hot")
                branch_plot_colorbar = True
            if "t" in branch_style:
                avgavgflow = np.mean(avgTot)
                lwidths = [2 * f / avgavgflow for f in avgTot]
            branch_label = "Branch flow"
        elif dcbranchtype == "capacity":
            cap = [data.branch.capacity[i] for i in range(len(data.branch.capacity)) if data.branch.type[i] == "dc"]
            branch_value = np.asarray(cap)
            maxcap = np.nanmax(branch_value)
            branch_colormap = plt.get_cmap("hot")
            branch_label = "Branch capacity"
            if filter_branch is None:
                # need an upper limit to avoid crash due to inf capacity
                filter_branch = [0, np.round(maxcap, -2) + 100]
            if "t" in branch_style:
                avgcap = np.mean(cap)
                lwidths = [2 * f / avgcap for f in cap]

        idx_from = data.dcBranchFromNodeIdx()  # empty
        idx_to = data.dcBranchToNodeIdx()  # empty
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)
        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        line_segments_dc = mpl.collections.LineCollection(ls, linewidths=lwidths, colors="black")
        # line_segments_dc = mpl.collections.LineCollection(
        #        ls, linewidths=2,colors='blue')
        if filter_branch is not None:
            line_segments_dc.set_clim(filter_branch)
        line_segments_dc.set_array(branch_value)
        ax.add_collection(line_segments_dc)

        # Nodes
        node_plot_colorbar = True
        if nodetype == "area":
            areas = data.node.area
            allareas = data.getAllAreas()
            # colours_co = cm.prism(np.linspace(0, 1, len(allareas)))
            node_label = "Node area"
            node_value = [allareas.index(c) for c in areas]
            node_colormap = cm.prism
            node_plot_colorbar = False
            # this is to get same colours as for branches:
            node_colormap.set_under("k")
            filter_node = [0, len(allareas)]
        elif nodetype == "nodalprice":
            avgprice = self.getAverageNodalPrices(timeMaxMin)
            node_label = "Nodal price"
            node_value = avgprice
            node_colormap = cm.jet
        elif nodetype == "lmp":
            node_value = self.getAverageNodalPrices(timeMaxMin)
            node_label = "Locational marginal price"
            node_colormap = cm.jet
        elif nodetype == "energybalance":
            avg_energybalance = self.getAverageEnergyBalance(timeMaxMin)
            node_label = "Nodal energy balance"
            node_value = avg_energybalance
            node_colormap = cm.hot
        elif nodetype == "loadshedding":
            node_value = self.getLoadsheddingPerNode(timeMaxMin)
            node_label = "Loadshedding"
            node_colormap = cm.hot
        else:
            node_value = "dimgray"
            node_colormap = cm.jet
            node_label = ""
            node_plot_colorbar = False

        x, y = m(data.node["lon"].tolist(), data.node["lat"].tolist())
        if nodetype == "lmp":
            sc = plt.hexbin(x, y, gridsize=20, C=node_value, cmap=node_colormap)
        else:
            sc = m.scatter(x, y, marker="o", c=node_value, cmap=node_colormap, zorder=2, s=dotsize)
        # sc.cmap.set_under('dimgray')
        # sc.cmap.set_over('dimgray')
        if filter_node is not None:
            sc.set_clim(filter_node[0], filter_node[1])

            # #TODO: Er dette nødvendig lenger, Harald?
            # #nodes with NAN nodal price plotted in gray:
            # for i in range(len(avgprice)):
            # if np.isnan(avgprice[i]):
            # m.scatter(x[i],y[i],c='dimgray',
            # zorder=2,s=dotsize)

        # NEW Colorbar for nodes
        # m. or plt.?
        if node_plot_colorbar:
            axcb2 = plt.colorbar(sc)
            axcb2.set_label(node_label)

        # NEW Colorbar for branch capacity
        if branch_plot_colorbar:
            axcb = plt.colorbar(line_segments_ac)
            axcb.set_label(branch_label)

        # Show names of nodes
        if show_node_labels:
            labels = data.node["id"]
            x1, x2, y1, y2 = plt.axis()
            offset_x = (x2 - x1) / 50
            for label, xpt, ypt in zip(labels, x, y):
                if xpt > x1 and xpt < x2 and ypt > y1 and ypt < y2:
                    plt.text(xpt + offset_x, ypt, label)

        if showTitle:
            plt.title("Nodes %s and branches %s" % (nodetype, branchtype))
        return

    def getEnergyMix(self, timeMaxMin=None, relative=False, showTitle=True, variable="energy"):
        """
        Get energy, generation capacity or spilled energy per area per type

        Parameters
        ----------
        timeMaxMin : list of two integers
            Time range, [min,max]
        relative : boolean
            Whether to plot absolute (false) or relative (true) values
        variable : string ("energy","capacity","spilled")
            Which variable to plot (default is energy production)
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        if variable == "energy":
            print("Getting energy output from all generators...")
            gen_output = self.db.getResultGeneratorPowerSum(timeMaxMin)
        elif variable == "capacity":
            gen_output = self.grid.generator.pmax
        elif variable == "spilled":
            gen_output = self.db.getResultGeneratorSpilledSums(timeMaxMin)
        else:
            print("Variable not valid")
            return

        df = self.grid.generator[["node", "type", "pmax"]].merge(
            self.grid.node[["id", "area"]], how="left", left_on="node", right_on="id"
        )
        df["VALUE"] = gen_output
        dfplot = df[["area", "type", "VALUE"]].groupby(["area", "type"]).sum()["VALUE"].unstack()

        if relative:
            dfplot = dfplot.mul(1 / dfplot.sum(axis=1), axis=0)

        return dfplot

    def plotEnergyMix(
        self, areas=None, timeMaxMin=None, relative=False, showTitle=True, variable="energy", gentypes=None
    ):
        """
        Plot energy, generation capacity or spilled energy as stacked bars

        Parameters
        ----------
        areas : list of strings
            Which areas to include, default=None means include all
        timeMaxMin : list of two integers
            Time range, [min,max]
        relative : boolean
            Whether to plot absolute (false) or relative (true) values
        variable : string ("energy","capacity","spilled")
            Which variable to plot (default is energy production)
        gentypes : list
            List of generator types to include. None gives all.
        """

        if timeMaxMin is None:
            timeMaxMin = [self.timerange[0], self.timerange[-1] + 1]

        dfplot = self.getEnergyMix(timeMaxMin=timeMaxMin, relative=relative, variable=variable)

        titles = {"energy": "Energy mix", "capacity": "Capacity mix", "spilled": "Energy spilled"}
        title = ""
        if variable in titles:
            title = titles[variable]
        if areas is None:
            areas = dfplot.index
        if gentypes is None:
            gentypes = dfplot.columns
        dfplot.loc[areas, gentypes].plot(kind="bar", stacked=True)
        plt.legend()
        handles, labels = plt.gca().get_legend_handles_labels()
        handles.reverse()
        labels.reverse()
        plt.legend(handles, labels, loc=2, bbox_to_anchor=(1.05, 1), borderaxespad=0.0)

        if showTitle:
            plt.title(title)
        return dfplot

    def plotTimeseriesColour(self, areas, value="nodalprice", filter_values=None):
        """
        Plot timeseries values with days on x-axis and hour of day on y-axis


        Parameters
        ----------
        areas : list of strings
            which areas to include, default=None means include all
        value : 'nodalprice' (default),
                'demand',
                'gen%<type1>%<type2>.. (where type=gentype)
            which times series value to plot

        Example: res.plotTimeseriescolour(['ES'],value='gen%solar_csp%wind')
        """

        p = {}
        pm = {}
        stepsperday = int(24 / self.grid.timeDelta)
        numdays = int(len(self.grid.timerange) / stepsperday)
        for a in areas:
            if value == "nodalprice":
                p[a] = self.getAreaPrices(area=a)
            elif value == "demand":
                # TODO: This is not generally correct. Should use
                # weighted average for all loads in area
                p[a] = self.grid.profiles["load_" + a]
            elif value[:3] == "gen":
                # value is now on form "gen_MA_hydro"
                strval = value.split("%")
                gens = self.grid.getGeneratorsPerAreaAndType()
                # genindx = gens[a][strval[1]]
                genindx = [i for s in strval[1:] for i in gens[a][s]]
                timerange = [self.timerange[0], self.timerange[-1] + 1]
                p[a] = self.db.getResultGeneratorPower(genindx, timerange)
            pm[a] = np.reshape(p[a], (numdays, stepsperday)).T

        # print("Plotting...")
        if not filter_values:
            vmin = min([min(p[a]) for a in areas])
            vmax = max([max(p[a]) for a in areas])
        else:
            vmin = filter_values[0]
            vmax = filter_values[1]
        num_areas = len(areas)
        fig, axes = plt.subplots(
            nrows=num_areas,
            ncols=1,
            figsize=((min(max(6, (len(self.grid.timerange) / 100)), 20)), (max(5, 1.5 * len(areas)))),
            sharex=True,
        )

        for n in range(num_areas):
            ax = plt.subplot(num_areas, 1, n + 1)
            ax.set_title(areas[n], x=-0.04, y=0.5, verticalalignment="center", horizontalalignment="right")
            plt.imshow(pm[areas[n]], vmin=vmin, vmax=vmax)

        fig.subplots_adjust(right=0.90)
        cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])
        # fig.colorbar(im,cax=cbar_ax)
        plt.colorbar(cax=cbar_ax)

    def plotRelativeLoadDistribution(
        self, show_node_labels=False, latlon=None, dotsize=40, draw_par_mer=False, colours=True, showTitle=True
    ):
        """
        Plots the relative input load distribution.

        Parameters
        ----------
        show_node_labels : boolean
            whether to show node names (true/false)
        latlon : list of four floats
            map area [lat_min, lon_min, lat_max, lon_max]
        draw_par_mer : boolean
            whether to draw parallels and meridians on map
        colours : boolean
            whether to draw the map in colours or black and white

        """
        # basemap is only used here, so to allow using powergama without
        # basemap installed, it is best to put import statement here.
        from collections import OrderedDict

        from mpl_toolkits.basemap import Basemap

        plt.figure()

        data = self.grid

        if latlon is None:
            lat_max = max(data.node.lat) + 1
            lat_min = min(data.node.lat) - 1
            lon_max = max(data.node.lon) + 1
            lon_min = min(data.node.lon) - 1
        else:
            lat_min = latlon[0]
            lon_min = latlon[1]
            lat_max = latlon[2]
            lon_max = latlon[3]

        # Use the average latitude as latitude of true scale
        lat_truescale = np.mean(data.node.lat)

        m = Basemap(
            resolution="l",
            projection="merc",
            lat_ts=lat_truescale,
            llcrnrlon=lon_min,
            llcrnrlat=lat_min,
            urcrnrlon=lon_max,
            urcrnrlat=lat_max,
            anchor="W",
        )

        # Draw coastlines, meridians and parallels.
        m.drawcoastlines()
        m.drawcountries(zorder=0)
        if colours:
            m.fillcontinents(color="coral", lake_color="aqua", zorder=0)
            m.drawmapboundary(fill_color="aqua")
        else:
            m.fillcontinents(zorder=0)
            m.drawmapboundary()

        if draw_par_mer:
            m.drawparallels(
                np.arange(_myround(lat_min, 10, "floor"), _myround(lat_max, 10, "ceil"), 10), labels=[1, 1, 0, 0]
            )

            m.drawmeridians(
                np.arange(_myround(lon_min, 10, "floor"), _myround(lon_max, 10, "ceil"), 10), labels=[0, 0, 0, 1]
            )

        # AC Branches
        idx_from = data.branchFromNodeIdx()
        idx_to = data.branchToNodeIdx()
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)

        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        line_segments_ac = mpl.collections.LineCollection(ls, linewidths=2, colors="k")

        ax = plt.axes()
        ax.add_collection(line_segments_ac)

        # DC Branches
        idx_from = data.dcBranchFromNodeIdx()
        idx_to = data.dcBranchToNodeIdx()
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)
        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        if colours:
            dcColour = "b"
        else:
            dcColour = "k"
        line_segments_dc = mpl.collections.LineCollection(ls, linewidths=2, colors=dcColour)

        ax.add_collection(line_segments_dc)

        # Loads
        x, y = m(data.node["lon"].tolist(), data.node["lat"].tolist())
        loadByNode = OrderedDict([(k, 0) for k in data.node["id"]])
        consNodes = data.consumer.node
        consValues = data.consumer.demand_avg

        for idx in range(len(consNodes)):
            loadByNode[consNodes[idx]] += consValues[idx]

        avgLoad = np.mean(consValues)
        relativeLoads = [dotsize * (myload[1] / avgLoad) for myload in loadByNode.items()]

        if colours:
            loadColour = "b"
        else:
            loadColour = "w"
        m.scatter(x, y, marker="o", c=loadColour, zorder=2, s=relativeLoads)

        # Show names of nodes
        if show_node_labels:
            labels = data.node["id"]
            x1, x2, y1, y2 = plt.axis()
            offset_x = (x2 - x1) / 50
            for label, xpt, ypt in zip(labels, x, y):
                if xpt > x1 and xpt < x2 and ypt > y1 and ypt < y2:
                    plt.text(xpt + offset_x, ypt, label)

        if showTitle:
            plt.title("Load distribution")

    def plotRelativeGenerationCapacity(
        self, tech, show_node_labels=False, latlon=None, dotsize=40, draw_par_mer=False, colours=True, showTitle=True
    ):
        """
        Plots the relative input generation capacity.

        Parameters
        ----------
        tech : string
            production technology to be plotted
        show_node_labels : boolean
            whether to show node names (true/false)
        latlon : list of four floats
            map area [lat_min, lon_min, lat_max, lon_max]
        draw_par_mer : boolean
            whether to draw parallels and meridians on map
        colours : boolean
            whether to draw the map in colours or black and white

        """

        # basemap is only used here, so to allow using powergama without
        # basemap installed, it is best to put import statement here.
        from collections import OrderedDict

        from mpl_toolkits.basemap import Basemap

        data = self.grid

        plt.figure()

        num_branches = data.branch.shape[0]

        genTypes = data.getAllGeneratorTypes()
        if tech not in genTypes:
            raise Exception(
                "No generators classified as " + tech + ".\nGenerator classifications: " + str(genTypes)[1:-1]
            )

        if latlon is None:
            lat_max = max(data.node.lat) + 1
            lat_min = min(data.node.lat) - 1
            lon_max = max(data.node.lon) + 1
            lon_min = min(data.node.lon) - 1
        else:
            lat_min = latlon[0]
            lon_min = latlon[1]
            lat_max = latlon[2]
            lon_max = latlon[3]

        # Use the average latitude as latitude of true scale
        lat_truescale = np.mean(data.node.lat)

        m = Basemap(
            resolution="l",
            projection="merc",
            lat_ts=lat_truescale,
            llcrnrlon=lon_min,
            llcrnrlat=lat_min,
            urcrnrlon=lon_max,
            urcrnrlat=lat_max,
            anchor="W",
        )

        # Draw coastlines, meridians and parallels.
        m.drawcoastlines()
        m.drawcountries(zorder=0)
        if colours:
            m.fillcontinents(color="coral", lake_color="aqua", zorder=0)
            m.drawmapboundary(fill_color="aqua")
        else:
            m.fillcontinents(zorder=0)
            m.drawmapboundary()

        if draw_par_mer:
            m.drawparallels(
                np.arange(_myround(lat_min, 10, "floor"), _myround(lat_max, 10, "ceil"), 10), labels=[1, 1, 0, 0]
            )

            m.drawmeridians(
                np.arange(_myround(lon_min, 10, "floor"), _myround(lon_max, 10, "ceil"), 10), labels=[0, 0, 0, 1]
            )

        lwidths = [2] * num_branches

        # AC Branches
        idx_from = data.branchFromNodeIdx()
        idx_to = data.branchToNodeIdx()
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)

        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        line_segments_ac = mpl.collections.LineCollection(ls, linewidths=lwidths, colors="k")

        ax = plt.axes()
        ax.add_collection(line_segments_ac)

        # DC Branches
        idx_from = data.dcBranchFromNodeIdx()
        idx_to = data.dcBranchToNodeIdx()
        branch_lat1 = [data.node.lat[i] for i in idx_from]
        branch_lon1 = [data.node.lon[i] for i in idx_from]
        branch_lat2 = [data.node.lat[i] for i in idx_to]
        branch_lon2 = [data.node.lon[i] for i in idx_to]

        x1, y1 = m(branch_lon1, branch_lat1)
        x2, y2 = m(branch_lon2, branch_lat2)
        ls = [[(x1[i], y1[i]), (x2[i], y2[i])] for i in range(len(x1))]
        if colours:
            dcColour = "b"
        else:
            dcColour = "k"
        line_segments_dc = mpl.collections.LineCollection(ls, linewidths=2, colors=dcColour)

        ax.add_collection(line_segments_dc)

        # Generators
        x, y = m(data.node["lon"].tolist(), data.node["lat"].tolist())
        generators = data.getGeneratorsPerType()[tech]
        genNode = data.generator.node
        genCapacity = data.generator.pmax
        capByNode = OrderedDict([(k, 0) for k in data.node.id])
        totCap = 0

        for gen in generators:
            capByNode[genNode[gen]] += genCapacity[gen]
            totCap += genCapacity[gen]

        avgCap = totCap / len(generators)
        relativeCap = [dotsize * (g[1] / avgCap) for g in capByNode.items()]

        if colours:
            loadColour = "b"
        else:
            loadColour = "w"
        m.scatter(x, y, marker="s", c=loadColour, zorder=2, s=relativeCap)

        # Show names of nodes
        if show_node_labels:
            labels = data.node.id
            x1, x2, y1, y2 = plt.axis()
            offset_x = (x2 - x1) / 50
            for label, xpt, ypt in zip(labels, x, y):
                if xpt > x1 and xpt < x2 and ypt > y1 and ypt < y2:
                    plt.text(xpt + offset_x, ypt, label)

        if showTitle:
            plt.title(tech + " capacity distribution")

    def plotGenerationScatter(self, area, tech=[], dotsize=300, annotations=True):
        """
        Scatter plot of generation capacity and correlation of inflow
        with load.

        Parameters
        ----------
        area : string
            area to plot
        tech : list of strings
            production technologies to plot. Empty list = all
        dotsize : integer
            adjust the size of scatterplots
        annotations: boolean
            whether to plot annotations
        """
        data = self.grid

        plt.figure()

        if len(tech) > 0:
            genTypes = data.getAllGeneratorTypes()
            for gt in tech:
                if gt not in genTypes:
                    raise Exception(
                        "No generators classified as " + gt + ".\n Generator classifications: " + str(genTypes)[1:-1]
                    )
        else:
            tech = data.getAllGeneratorTypes()

        genByType = data.getGeneratorsPerAreaAndType()[area]
        detailedGen = {}
        for gt in genByType.keys():
            if gt in tech:
                detailedGen[gt] = {}
                genIdx = genByType[gt]
                for i in genIdx:
                    capacity = data.generator.pmax[i]
                    infProfile = data.generator.inflow_ref[i]
                    if infProfile in detailedGen[gt]:
                        detailedGen[gt][infProfile] += capacity
                    else:
                        detailedGen[gt][infProfile] = capacity

        cons = data.consumer.demand_ref[data.getConsumersPerArea()[area][0]]
        demandProfile = data.profiles[cons]
        x, y, label, size, colCoor, tickLabel = [], [], [], [], [], []
        mainCount = 0
        for gt in detailedGen.keys():
            tickLabel.append(gt)
            subCount = 0
            for infProf in detailedGen[gt].keys():
                if infProf == "const":
                    y.append(1)
                else:
                    y.append(np.corrcoef(data.profiles[infProf], demandProfile)[0, 1])
                x.append(mainCount)
                label.append(infProf)
                size.append(detailedGen[gt][infProf])
                subCount += 1
            mainCount += 1

        avgSize = np.mean(size)
        size = [dotsize * s / avgSize for s in size]
        colCoor = [c / mainCount for c in x]
        colours = mpl.cm.hsv(colCoor)
        plt.scatter(x, y, s=size, c=colours, alpha=0.5)
        if annotations:
            for label, x, y in zip(label, x, y):
                plt.annotate(
                    label,
                    xy=(x, y),
                    xytext=(20, 0),
                    textcoords="offset points",
                    ha="left",
                    va="bottom",
                    arrowprops=dict(arrowstyle="-", connectionstyle="arc3,rad=0"),
                )

        plt.ylabel("Correlation coefficient of inflow to load")
        plt.xlim(xmin=-0.5, xmax=mainCount + 0.5)
        plt.xticks(range(mainCount + 1), tickLabel)


def _myround(x, base=1, method="round"):
    """Round to nearest multiple of base"""
    if method == "round":
        return int(base * round(float(x) / base))
    elif method == "floor":
        return int(base * math.floor(float(x) / base))
    elif method == "ceil":
        return int(base * math.ceil(float(x) / base))
    else:
        raise



import pathlib
import time

import numpy as np

from . import failure_situation, specify_storage


def create_fault_scenarios(gridmodel_base, fault_path, seed_list, fault_spec):
    """Create files specifying fault scenario, saved in specified folder

    gridmodel_base : powergama.GridData - the grid model used as base
    fault_path : pathlib.Path where fault scenario files will be stored
    seed_list : list of random seeds used for fault scenario creation. The lengh of this list
               determines how many scenarios are created.
    fault_spec : FaultSpec - specificatino for faults
    """

    fault_duration = fault_spec.generator_fault_duration

    # Make folder if it does not exist:
    pathlib.Path(fault_path).mkdir(exist_ok=True)

    N_steps = gridmodel_base.timerange[-1]

    for seed in seed_list:
        switches_off = failure_situation.build_switches_off_array(
            N_steps=N_steps, data=gridmodel_base, fault_spec=fault_spec, seed=seed
        )
        fault_situation_list = failure_situation.build_off_states_list(
            switches_off, N_steps=N_steps, generator_fault_duration=fault_duration
        )
        print(seed, len(fault_situation_list) * fault_duration / N_steps)

        failure_situation.write_fault_sits_to_file(fault_path / f"fault_scenario_{seed}.txt", fault_situation_list)
    print(f"Finished writing {len(seed_list)} scenarios to folder")
    print(fault_path)


def run_fault_simulation(gridmodel_base, full_profiles, fault_scenario_file, failure_dir, db_base, solver):
    """Run simulation with faults, as specified in fault scenario file

    gridmodel_base : powergama.GridData - base grid model
    full_profiles : FullProfiles
    fault_scenario_file : name of fault scenario file
    failure_dir : pathlib.Path - where to save sql file with fault scenario results
    db_base : sqlite file with base model results
    solver : name of solver to use (glpk, cbc, gurobi_persistent, gurobi,...)
    """

    # Load base results and model
    base_case_storage = specify_storage.load_storage_states_from_res(db_base)
    base_case_flexload = specify_storage.load_flexload_states_from_res(db_base)

    N_steps = gridmodel_base.timerange[-1]
    print(f"N_steps: {N_steps}")

    # Read in all fault situations from fault scenario file
    fault_situation_list = failure_situation.read_fault_sits_from_file(fault_scenario_file)

    print("Number of fault situations: %s" % len(fault_situation_list))
    # print('Fraction of hours to be resimulated: %s' %(len(fault_situation_list)*FAULT_DUR/N_steps))
    start_all = time.time()

    # For each fault situation in our timeline, run the simulation
    for ii in range(len(fault_situation_list)):
        if ii % 1000 == 0:
            print(f"{ii}/{len(fault_situation_list)}")
        failure_situation.collect_res_failure_situation(
            fault_situation_list[ii],
            full_profiles=full_profiles,
            gridmodel_base=gridmodel_base,
            failure_dir=failure_dir / f"failure_{ii}",
            base_case_storage=base_case_storage,
            base_case_flexload=base_case_flexload,
            solver=solver,
        )
    end_all = time.time()
    print("Total simulation time: %s" % (np.round(end_all - start_all, 2)))



'''
Mehtods to create seasonal and daily filters of hourly timeseries typically
used in PowerGAMA.
'''
from __future__ import division
import time
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.ticker import AutoMinorLocator
from math import pi, sqrt, exp, sin


def gauss(n=11,sigma=1):
    r = range(-int(n/2),int(n/2)+1)
    return [1 / (sigma * sqrt(2*pi)) * exp(-float(x)**2/(2*sigma**2)) for x in r]


def sinc(n=11,T=1):
    r = range(-int(n/2),int(n/2)+1)
    return [ ( sin(pi * ((float(x) + 0.000000000001 ) / T) )
              / (pi * (float(x) + 0.000000000001 ) ) ) for x in r]


def SeasonFilter(HourlyTimeSeries,FilterPeriodInMonths=4,
    PlotFilter=False, Quick=False):
    '''Create seasonal filters

    Parameters
    ----------
    HourlyTimeSeries : pandas.dataframe
        orignial time series, should be one year
    FilterPeriodInMonths : float
        filter period
    PlotFilter : boolean
        wheter to make a plot
    Quick : boolean
        whether to make a quick
    '''

    time_filter_start = time.time()

    HoursPerYear = HourlyTimeSeries.shape[0]
    HoursPerMonth = HoursPerYear/12
    FilterLengthInYears = 4*FilterPeriodInMonths

    if Quick:
        FilterLengthInYears = FilterLengthInYears/4

    FilterLengthInHours = FilterLengthInYears*HoursPerYear


    if float(FilterLengthInYears).is_integer():
        FilterLengthInYears = int(FilterLengthInYears)
    else:
        print('Problem')
        raise

    filter_T = HoursPerMonth*FilterPeriodInMonths/2
    filter_n = FilterLengthInHours-1
    filter_offset = (filter_n/2)-0.5

    IdealFilter = pd.Series(sinc(n=filter_n, T=filter_T))
    IdealFilter.index = IdealFilter.index.values - filter_offset

    sigma = (FilterLengthInHours)/10

    if Quick:
        sigma = sigma*1.5

    window = pd.Series(gauss(n=filter_n, sigma=sigma))
    window.index = window.index.values - filter_offset

    window_normalised = pd.Series(window.values / window.loc[0])
    window_normalised.index = window_normalised.index.values - filter_offset

    window_plotting = pd.Series(window_normalised.values * IdealFilter.loc[0])
    window_plotting.index = window_plotting.index.values - filter_offset

    CroppedlFilter = pd.Series(IdealFilter.values * window_normalised.values)
    CroppedlFilter.index = CroppedlFilter.index.values - filter_offset

    if PlotFilter:
        plt.figure()
        IdealFilter.plot.line(figsize=(8 , 4.5), label='Ideal filter')
        window_plotting.plot.line(label='Window')
        CroppedlFilter.plot.line(label='Cropped Filter')
        plt.legend()

    #Sum Numbers for Quality Check

    Ideal_filter_integral_error = 1 - IdealFilter.sum()
    Ideal_filter_Cut_off = IdealFilter.iloc[0]/IdealFilter.loc[0]

    Window_integral_error = 1 - window.sum()
    Window_Cut_off = window_normalised.iloc[0]/window_normalised.loc[0]

    Combined_filter_integral_error = 1 - CroppedlFilter.sum()
    Combined_filter_Cut_off = CroppedlFilter.iloc[0]/CroppedlFilter.loc[0]

    HourlyTimeSeriesExtended = HourlyTimeSeries

    for counterfilterlength in range(FilterLengthInYears):
        HourlyTimeSeriesExtended = HourlyTimeSeriesExtended.append(
            HourlyTimeSeries, ignore_index=True)

    HourlyTimeSeriesExtended.index = (
        HourlyTimeSeriesExtended.index.values - (FilterLengthInHours/2))
    HourlyTimeSeriesFiltered = HourlyTimeSeries*0

    for counter_hours in range(HoursPerYear):
        temp1 = HourlyTimeSeriesExtended.iloc[:,:].mul(CroppedlFilter, axis=0)
        temp2 = pd.DataFrame(temp1.sum(axis=0)).T

        HourlyTimeSeriesFiltered.iloc[counter_hours, :] = temp2.iloc[0,:]
        HourlyTimeSeriesExtended.index = HourlyTimeSeriesExtended.index.values - 1

    time_filter_end = time.time()
    return HourlyTimeSeriesFiltered


def SeasonPlot(HourlyTimeSeries,fname=None,ylabel='',title=None):

    plt.rcParams.update({'font.size': 22})
    list_monate_kurz = ['Jan.','Feb.','Mar.','Apr.','Mai','Jun.',
                        'Jul.','Aug.','Sep.','Okt.','Nov.','Dec.']
    HoursPerYear = HourlyTimeSeries.shape[0]
    HoursPerMonth = HoursPerYear/12

    list_legend = list(HourlyTimeSeries.columns)
    list_legend.insert(0,'_')

    flatline = pd.DataFrame([1 for i in range(HoursPerYear)])

    fig4 = plt.figure(figsize=(16 , 9))
    ax4 = fig4.add_subplot(111)
    flatline.plot(ax=ax4, linewidth=2, color=['grey'])

    HourlyTimeSeries.plot(ax=ax4, linewidth=2,
        color=['r','y','b','c','black'],style=['-','-','-','-','--'])

    ax4.legend(list_legend,loc="lower left")

    if title is not None:
        plt.title(title)
    plt.xlabel("Month")
    plt.ylabel(ylabel)

    plt.xlim(0,HoursPerYear)

    labelpositions = np.arange(0.5*HoursPerMonth,(12.5)*HoursPerMonth,1*HoursPerMonth)
    plt.xticks(labelpositions, list_monate_kurz)

    minor_locator = AutoMinorLocator(2)
    plt.gca().xaxis.set_minor_locator(minor_locator)
    plt.grid(which='minor')

    if fname is not None:
        plt.savefig(fname,bbox_inches = 'tight',dpi=300)



def DayFilter(HourlyTimeSeries):
    '''Filter timeseries

    Parameters
    ----------
    HourlyTimeSeries : pandas.DataFrame
        time series to be filtered. Should be hourly time steps
    '''

    NumberOfSeries = HourlyTimeSeries.shape[1]
    NumberOfHours = HourlyTimeSeries.shape[0]
    HoursPerDay = 24
    NumberOfDays = NumberOfHours/HoursPerDay

    if float(NumberOfDays).is_integer():
        NumberOfDays = int(NumberOfDays)
    else:
        print('Problem')
        raise

    Matrix = np.zeros(shape=(NumberOfDays,HoursPerDay,NumberOfSeries))

    for counter_hours in range(NumberOfHours):
        counter_hoursinday = int(counter_hours % HoursPerDay)
        counter_days = int((counter_hours - counter_hoursinday) / HoursPerDay)
        Matrix[counter_days,counter_hoursinday,:]   = HourlyTimeSeries.iloc[counter_hours,:]

    ADHTS = pd.DataFrame(Matrix.mean(axis=0))
    ADHTS.columns = HourlyTimeSeries.columns
    return ADHTS


def DayPlot(ADHTS,fname=None,ylabel='',title=None):
    '''Plot the output from DayFilter'''

    plt.rcParams.update({'font.size': 22})

    HoursPerDay = 24
    list_legend = list(ADHTS.columns)
    list_legend.insert(0,'_')

    ADHTS = ADHTS.append(pd.DataFrame(ADHTS.iloc[-1,:]).T, ignore_index=True)

    flatline = pd.DataFrame([1 for i in range(HoursPerDay+1)])

    fig4 = plt.figure(figsize=(16 , 9))
    ax4 = fig4.add_subplot(111)
    flatline.plot(ax=ax4, linewidth=2, color=['grey'])

    ADHTS.plot(ax=ax4, drawstyle='steps-post', linewidth=2,
        color=['r','y','b','c','black'],style=['-','-','-','-','--'])

    ax4.legend(list_legend) # ,loc="upper left"

    if title is not None:
        plt.title(title)
    plt.xlabel("Hour")
    plt.ylabel(ylabel)

    plt.xlim(0,HoursPerDay)

    labelpositions = np.arange(0.5,HoursPerDay+0.5,1)
    plt.xticks(labelpositions, range(1,HoursPerDay+1))

    minor_locator = AutoMinorLocator(2)
    plt.gca().xaxis.set_minor_locator(minor_locator)
    plt.grid(which='minor')

    if fname is not None:
        plt.savefig(fname,bbox_inches = 'tight',dpi=300)



"""
This file has functions to set up fault simulations and saving and reading from file, as well as
collecting results for a fault situation.
"""

import pathlib

import numpy as np

from .full_profiles import get_gridmodel_failure
from .specify_storage import run_failure_case_LpFaultProblem


class FaultSpec:
    """Specification of faults"""

    def __init__(self, spec_generators=None, spec_branches=None):
        """Specify fault

        spec_generators : dict {fault_rate: float, fault_duration: float, fault_sizes : dict {type:value, None:default}}

        """
        if spec_generators:
            self.generator_fault_rate = spec_generators["fault_rate"]
            self.generator_fault_duration = spec_generators["fault_duration"]
            self.generator_fault_size = spec_generators["fault_sizes"]
        else:
            # default values if none specified:
            self.generator_fault_rate = 0.5665 / 8760
            self.generator_fault_duration = 21
            self.generator_fault_sizes = {None: 1000}

        # TODO: specify for branches

    def get_unit_size(self, gen_type):
        if gen_type in self.generator_fault_size:
            return self.generator_fault_size[gen_type]
        else:
            return self.generator_fault_size[None]


def collect_res_failure_situation(
    fail_situation, full_profiles, gridmodel_base, failure_dir, base_case_storage, base_case_flexload, solver
):
    """For a specific failure situation, run the simulation for the specified duration

    fail_situtaion: tuple - (start, end, failed_generators, failed_branches)
    failure_dir : pathlib.Path - where files are
    """
    start, end, failed_generators, failed_branches = fail_situation

    def get_gridmodel_specific_failure(unused, timerange):
        return get_gridmodel_failure(
            full_profiles,
            gridmodel_base,
            timerange,
            failed_generators=failed_generators,
            failed_branches=failed_branches,
        )

    test_timesteps = [start]
    pathlib.Path(failure_dir).mkdir(exist_ok=True)
    run_failure_case_LpFaultProblem(
        failure_dir,
        None,
        base_case_storage,
        base_case_flexload,
        get_gridmodel_specific_failure,
        test_timesteps,
        num_steps=end - start,
        solver=solver,
        lossmethod=0,
    )


def build_switches_off_array(N_steps, data, fault_spec, seed=None, verbose=False, exclude_nodes=None):
    """

    N_steps : int
    data : powergama.GridModel object - grid model
    fault_spec : FaultSpec object - fault situation specification
    """
    fault_rate = fault_spec.generator_fault_rate

    if seed is not None:
        np.random.seed(seed)
    N_gen = data.numGenerators()
    # Array specifying whether a generator turns off at that time
    # Given in the percentage of switch-off, so a value between 0 and 1
    switches_off = np.zeros((N_gen, N_steps))

    # Iterate through generators and sample when and how much generators fail
    for gg in range(N_gen):
        gen = data.generator.iloc[gg]
        if exclude_nodes and (gen.node in exclude_nodes):
            # amount_off = 0  # We don't simulate generator faults at these nodes
            continue
        unit_size = fault_spec.get_unit_size(gen.type)
        if gen.pmax <= unit_size:
            # If there is only one we check whether it is on or off
            switches_off[gg, :] = np.random.binomial(n=1, p=fault_rate, size=N_steps)
        else:
            # We have num_gens generators, of which num_full_gens are of unit_size and the
            # potential remaining one is of remaining size.
            num_gens = np.ceil(gen.pmax / unit_size)
            if num_gens * unit_size > gen.pmax:
                num_full_gens = num_gens - 1
                remaining = gen.pmax - num_full_gens * unit_size
            else:
                num_full_gens = num_gens
                remaining = 0
            # Sample whether the generators are on or off
            # Done independently for each generator and each timestep
            switches_off_full = np.random.binomial(n=num_full_gens, p=fault_rate, size=N_steps)
            switches_off_rem = np.random.binomial(n=1, p=fault_rate, size=N_steps)
            if verbose and (switches_off_full.any() or switches_off_rem.any()):
                print(gen)
                print(switches_off_full, switches_off_rem)
            switches_off[gg, :] = switches_off_full * unit_size / gen.pmax + switches_off_rem * remaining / gen.pmax
    return switches_off


def build_off_states_list(switches_off, N_steps, generator_fault_duration):
    # List of tuples of (start, end, generators_fail, branches_fail)
    # TODO: branches_fail
    # generators_fail: list of tuples of (gg_idx, amount_off)
    # branches_fail: TODO
    fault_situation_list = []
    tt = 0

    while tt < N_steps:
        if switches_off[:, tt].any():
            start = tt
            end = min(tt + generator_fault_duration, N_steps)  # not inclusive
            # TODO: N_steps, or plus or minus one?
            generators_amount = []
            # Find any other ones that start within this area and also set them to start at tt
            to_turn_off = np.nonzero(switches_off[:, start:end].any(1))[0]
            # Set any that start at tt to be off until tt + duration
            for gg in to_turn_off:
                amount_off = np.max(switches_off[gg, start:end])
                # HACK Need to add int(...) and float(...) to avoid getting np.int64(...) in created files
                # Code should be improved to avoid such things
                generators_amount.append((int(gg), float(amount_off)))
            fault_situation_list.append((start, end, generators_amount, []))
            # jump to tt + duration
            tt += generator_fault_duration
        else:
            tt += 1
    return fault_situation_list


def build_off_states_array(fault_situation_list, data, N_steps):
    N_gen = data.numGenerators()
    off_states = np.zeros((N_gen, N_steps))
    for fault in fault_situation_list:
        start, end, failed_generators, failed_branches = fault

        for gg, amount in failed_generators:
            off_states[gg, start:end] = amount
    # TODO: failed_branches
    return off_states


def write_fault_sits_to_file(filename, fault_situation_list):
    ff = open(filename, "w")
    for fault in fault_situation_list:
        start, end, generators, lines = fault
        ff.write(f"{start}, {end}\n")
        for gg_tup in generators:
            ff.write(f"{gg_tup}, ")
        ff.write("\n")
        for ll_tup in lines:
            ff.write(f"{ll_tup}, ")
        ff.write("\n")
        ff.write("\n")
    ff.close()


def parse_comps(gg_str):
    # Read in a sequence of tuples in a string and return a list of tuples (int, float)
    gg_list = []
    if not (gg_str.split("), ")[-1] == "\n"):
        raise Exception("Error in parse_comps")
    # Get a list of the tuple strings
    gg_tup_list = gg_str.split("), ")[:-1]
    # For each tuple string, parse it into a tuple of (int, float)
    for ii in range(len(gg_tup_list)):
        gg_tup_ex = gg_tup_list[ii]
        gg_idx_ex, gg_frac = gg_tup_ex.split(", ")
        gg_idx = gg_idx_ex.split("(")[1]
        gg_list.append((int(gg_idx), float(gg_frac)))
    return gg_list


def read_fault_sits_from_file(filename):
    ff = open(filename, "r")
    rows = ff.readlines()
    ff.close()

    new_fault_list = []
    max_ii = len(rows) - 1
    ii = 0
    while True:
        start, end_ex = rows[ii].split(", ")
        end = end_ex.split("\n")[0]
        gg_list = parse_comps(rows[ii + 1])
        ll_list = parse_comps(rows[ii + 2])
        if not (rows[ii + 3] == "\n"):
            raise Exception("There should be an empty line between each fault situation")
        new_fault_list.append((int(start), int(end), gg_list, ll_list))
        ii += 4
        if ii > max_ii:
            break

    return new_fault_list
